#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#

import nanaimo
import nanaimo.version


def test_arg_version(run_nait) -> None:  # type: ignore
    """
    Verify --version argument contract.
    """
    expected = nanaimo.version.__version__
    assert run_nait(['--version']).stdout.decode('utf-8').startswith(expected)


def test_help(run_nait) -> None:  # type: ignore
    """
    Verify --help doesn't explode.
    """
    result = run_nait(['gtest_over_jlink', '--help']).stdout.decode('utf-8')
    assert len(result) > 4
    print(result)


def test_nanaimo_bar(run_nait) -> None:  # type: ignore
    """
    Eat your dessert!
    """
    result = run_nait(['--log-level', 'VERBOSE_DEBUG', 'nanaimo_bar']).stdout.decode('utf-8')
    print(result)



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
#                                       (@@@@%%%%%%%%%&@@&.
#                              /%&&%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@@(
#                              *@&%%%%%%%%%&&%%%%%%%%%%%%%%%%%%&&&%%%%%%%
#                               @   @@@(@@@@%%%%%%%%%%%%%%%%&@@&* @@@   .
#                               ,   .        .  .@@@&                   /
#                                .       .                              *
#                               @@              .                       @
#                              @&&&&&&@. .    .                     *@%&@
#                              &&&&&&&&&&&&&&&&@@        *@@############@
#                     *&/ @@ #&&&&&&&&&&&&&&&&&&&&@  ###################*
#                              @&&&&&&&&&&&&&&&&&&##################@
#                                 %@&&&&&&&&&&&&&&################@
#                                        @&&&&&&&&&&%#######&@%
#  nanaimo                                   (@&&&&####@@*
#
import argparse
import asyncio
import logging
import textwrap
import typing

import argcomplete

import nanaimo
import nanaimo.fixtures
import nanaimo.config


class CreateAndGatherFunctor:
    """
    Stores the type, manager, and loop then uses these to instantiate
    and invoke a Fixture if the given default is selected.
    Returns the result-code of the artifacts.
    """

    def __init__(self,
                 fixture_type: typing.Type[nanaimo.fixtures.Fixture],
                 manager: nanaimo.fixtures.FixtureManager,
                 loop: asyncio.AbstractEventLoop):
        self._fixture_type = fixture_type
        self._manager = manager
        self._loop = loop
        self._logger = logging.getLogger(fixture_type.get_canonical_name())

    async def __call__(self, args: nanaimo.Namespace, gather_timeout_seconds: typing.Optional[float] = None) -> int:
        fixture = self._fixture_type(self._manager,
                                     args,
                                     loop=self._loop,
                                     gather_timeout_seconds=gather_timeout_seconds)
        artifacts = await fixture.gather()
        artifacts.dump(self._logger)
        return int(artifacts)


def _auto_brief(documented: typing.Any, default_brief: str = '') -> str:
    if documented is None or documented.__doc__ is None:
        return default_brief
    lines = documented.__doc__.strip().split('\n')  # type: typing.List[str]
    if len(lines) > 1:
        return '{}…'.format(lines[0])
    else:
        return lines[0]


def _visit_argparse(manager: nanaimo.fixtures.FixtureManager,
                    subparsers: argparse._SubParsersAction,
                    loop: asyncio.AbstractEventLoop,
                    defaults: typing.Optional[nanaimo.config.ArgumentDefaults]) -> None:
    for fixture_type in manager.fixture_types():
        subparser = subparsers.add_parser(fixture_type.get_canonical_name(),
                                          help=_auto_brief(fixture_type))  # type: 'argparse.ArgumentParser'
        fixture_type.on_visit_test_arguments(nanaimo.Arguments(subparser, defaults, fixture_type.get_argument_prefix()))
        subparser.set_defaults(func=CreateAndGatherFunctor(fixture_type, manager, loop))


def _make_parser(loop: typing.Optional[asyncio.AbstractEventLoop] = None,
                 defaults: typing.Optional[nanaimo.config.ArgumentDefaults] = None) -> argparse.ArgumentParser:
    """
    Defines the command-line interface. Provided as a separate factory method to
    support sphinx-argparse documentation.
    """

    epilog = '''**Example Usage**::

    python -m nanaimo -vv nanaimo_bar

----
'''

    if defaults is None:
        defaults = nanaimo.config.ArgumentDefaults()

    parser = argparse.ArgumentParser(
        description='Run tests against hardware.',
        epilog=epilog,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    from nanaimo.version import __version__

    parser.add_argument('--version', action='version', version=__version__)

    parser.add_argument('--rcfile', help='Path to a default values configuration file.'
                                         'See nanaimo.Namespace for details.')
    parser.add_argument('--log-level', choices=['WARNING', 'INFO', 'DEBUG', 'VERBOSE_DEBUG'],
                        help='python logging level.')

    parser.add_argument('--gather-timeout-seconds',
                        type=float,
                        help=textwrap.dedent('''
                            A gather timeout in fractional seconds to use for all fixtures.
                            If not provided then Fixture.gather will not timeout.''').lstrip())

    subparsers = parser.add_subparsers(dest='fixture', help='Available fixtures.')

    pm = nanaimo.fixtures.PluggyFixtureManager()

    _visit_argparse(pm, subparsers, loop, defaults)  # type: ignore

    return parser


def _setup_logging(args: nanaimo.Namespace) -> None:
    fmt = '%(asctime)s %(levelname)s %(name)s: %(message)s'
    if args.log_level == 'WARNING':
        level = logging.WARNING
    elif args.log_level == 'DEBUG':
        level = logging.DEBUG
    elif args.log_level == 'VERBOSE_DEBUG':
        level = logging.DEBUG
        logging.getLogger('asyncio').setLevel(logging.DEBUG)
    else:
        level = logging.INFO
    logging.basicConfig(level=level, format=fmt, datefmt='%Y-%m-%d %H:%M:%S')
    if level <= logging.INFO:
        logging.getLogger(__name__).info('Nanaimo logging is configured.')


def main() -> int:
    """
    CLI entrypoint for running Nanaimo tests.
    """

    loop = asyncio.get_event_loop()
    defaults = nanaimo.config.ArgumentDefaults.create_defaults_with_early_rc_config()

    parser = _make_parser(loop, defaults)
    argcomplete.autocomplete(parser)
    args = parser.parse_args()
    defaults.set_args(args)

    args_ns = nanaimo.Namespace(args, defaults, allow_none_values=False)

    _setup_logging(args_ns)

    if hasattr(args, 'func'):
        result = loop.run_until_complete(args.func(args_ns, args_ns.gather_timeout_seconds))
        try:
            return int(result)
        except ValueError:
            logging.getLogger(__name__).error('Nanaimo tests must return an int result!')
            raise
    else:
        parser.print_usage()
        return -1



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
#                                       (@@@@%%%%%%%%%&@@&.
#                              /%&&%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@@(
#                              *@&%%%%%%%%%&&%%%%%%%%%%%%%%%%%%&&&%%%%%%%
#                               @   @@@(@@@@%%%%%%%%%%%%%%%%&@@&* @@@   .
#                               ,   .        .  .@@@&                   /
#                                .       .                              *
#                               @@              .                       @
#                              @&&&&&&@. .    .                     *@%&@
#                              &&&&&&&&&&&&&&&&@@        *@@############@
#                     *&/ @@ #&&&&&&&&&&&&&&&&&&&&@  ###################*
#                              @&&&&&&&&&&&&&&&&&&##################@
#                                 %@&&&&&&&&&&&&&&################@
#                                        @&&&&&&&&&&%#######&@%
#  nanaimo                                   (@&&&&####@@*
#
"""
Connections are built-in async abstractions using standard communication
protocols like UART, I2C, CAN, TCP/IP, etc. :class:`Instrument` and :class:`nanaimo.fixtures.Fixture`
classes use connections to bind to physical hardware.
"""
import asyncio
import codecs
import queue
import typing

import nanaimo


class TimestampedLine(str):
    """
    A line of text with an associated timestamp. This is a subclass of string so the
    object may be treated as a string without conversion.
    """

    @classmethod
    def create(cls, line_text: object, timestamp_seconds: float) -> 'TimestampedLine':
        timestamped = TimestampedLine(line_text)
        timestamped._timestamp_seconds = timestamp_seconds
        return timestamped

    def __init__(self, line_text: object):
        self._timestamp_seconds = 0.0

    @property
    def timestamp_seconds(self) -> float:
        """
        The timestamp for the line in fractional seconds. For example, this would be the time
        the line of text was received when this type is returned for a getter.
        """
        return self._timestamp_seconds


class AbstractSerial:
    """
    Abstract base class for a serial communication channel.
    """

    @classmethod
    def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
        arguments.add_argument('--port',
                               help='The port to monitor.')

        arguments.add_argument('--port-speed', help='the speed of the port (e.g. baud rate for serial ports).')


class AbstractAsyncSerial(AbstractSerial):
    """
    Abstract base class for a serial communication channel that provides
    asynchronous methods.
    """

    def __init__(self, loop: typing.Optional[asyncio.AbstractEventLoop] = None) -> None:
        self._loop = (loop if loop is not None else asyncio.get_event_loop())
        self._read_buffer = queue.Queue()  # type: queue.Queue[TimestampedLine]
        self._write_buffer = queue.Queue()  # type: queue.Queue[str]
        self._rx_decoder = codecs.getincrementaldecoder('UTF-8')('replace')
        self._tx_encoder = codecs.getincrementalencoder('UTF-8')('replace')
        self._queues_are_running = True
        self._rx_buffer_overflows = 0

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    @property
    def rx_buffer_overflows(self) -> int:
        return self._rx_buffer_overflows

    def stop(self) -> None:
        self._queues_are_running = False

    def time(self) -> float:
        """
        Get the current, monotonic time, in fractional seconds, using the same
        clock used for receive timestamps.
        """
        return self._loop.time()

    # +-----------------------------------------------------------------------+
    # | ASYNC OPERATIONS
    # +-----------------------------------------------------------------------+
    async def get_line(self, timeout_seconds: float = 0) -> TimestampedLine:
        """
        Get a line of text.

        :param float timeout_seconds: Time in fractional seconds to wait for input.
        :returns: A line of text with the time it was received at.
        :rtype: TimestampedLine
        :raises asyncio.TimeoutError: If a full line of text was not received within
            the specified timeout period.
        """
        start_time = self.time()
        while True:
            try:
                return self._read_buffer.get_nowait()
            except queue.Empty:
                if not self._queues_are_running:
                    raise
                if timeout_seconds > 0 and self.time() - start_time > timeout_seconds:
                    raise asyncio.TimeoutError()
                await asyncio.sleep(0.001)

    async def put_line(self, input_line: str, timeout_seconds: float = 0) -> float:
        """
        Put a line of text to the serial device.

        :param str input_line: The line to put.
        :param float timeout_seconds: Fractional seconds to block for if the input buffer is full. If the buffer does
            not become available within this time then :class:`asyncio.TimeoutError` is raised. Use 0 to block
            forever.
        :return: The monotonic system time that the line was put into the serial buffers at (see :meth:`time`).
        :raises asyncio.TimeoutError: If an input buffer did not become available within the specified timeout.
        """
        start_time = self.time()
        while self._queues_are_running:
            try:
                start_of_put = self.time()
                self._write_buffer.put_nowait(input_line)
                return start_of_put
            except queue.Full:
                if timeout_seconds > 0 and self.time() - start_time > timeout_seconds:
                    raise asyncio.TimeoutError()
                await asyncio.sleep(0.001)
        return self.time()



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
#                                       (@@@@%%%%%%%%%&@@&.
#                              /%&&%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@@(
#                              *@&%%%%%%%%%&&%%%%%%%%%%%%%%%%%%&&&%%%%%%%
#                               @   @@@(@@@@%%%%%%%%%%%%%%%%&@@&* @@@   .
#                               ,   .        .  .@@@&                   /
#                                .       .                              *
#                               @@              .                       @
#                              @&&&&&&@. .    .                     *@%&@
#                              &&&&&&&&&&&&&&&&@@        *@@############@
#                     *&/ @@ #&&&&&&&&&&&&&&&&&&&&@  ###################*
#                              @&&&&&&&&&&&&&&&&&&##################@
#                                 %@&&&&&&&&&&&&&&################@
#                                        @&&&&&&&&&&%#######&@%
#  nanaimo                                   (@&&&&####@@*
#
"""
Much of what Nanaimo does is to allow for intuitive, intelligent, and adaptable configuration of fixtures. This module
contains internal implementations used by public types to facilitate this functionality.
"""
import argparse
import configparser
import logging
import os
import pathlib
import typing
import weakref


class ArgumentDefaults:
    """
    Manages default values for Nanaimo arguments. Use in conjunction with :class:`nanaimo.Arguments` and
    :class:`nanaimo.Namespace` to populate default argument values when adding arguments to pytest or argparse
    and to replace missing attributes with values from configuration.

    When using the Nanaimo CLI or pytest plugin you don't need to worry about this object as these modules
    wire up the defaults, arguments, and namespaces for you.
    """

    _default_read_locations = ['~/nanaimo.cfg', '/etc/nanaimo.cfg', 'setup.cfg']
    """
    These are all the locations that Nanaimo configuration files will be searched for.
    Configuration merge rules are as defined by :meth:`configparser.ConfigParser.read`.
    """

    @classmethod
    def create_defaults_with_early_rc_config(cls) -> 'ArgumentDefaults':
        '''
        A special factory method that creates a :class:`ArgumentDefaults` instance pulling
        the value of ``--rcfile`` directly from :data:`sys.argv`. This allows defaults to
        be pulled from a config file before argument parsing is performed.
        '''
        import sys

        def args() -> None:
            pass

        for x in range(0, len(sys.argv) - 1):
            if sys.argv[x] == '--rcfile':
                setattr(args, 'rcfile', sys.argv[x+1])
                break

        return ArgumentDefaults(args)

    def __init__(self, args: typing.Optional[typing.Any] = None) -> None:
        self._env_variable_index = weakref.WeakKeyDictionary()  # type: weakref.WeakKeyDictionary
        self._configparser = configparser.ConfigParser(interpolation=configparser.ExtendedInterpolation())
        self._logger = logging.getLogger(__name__)
        if args is not None:
            self.set_args(args)
        else:
            self.set_args(dict())

    def set_args(self, args: typing.Any) -> None:
        if hasattr(args, 'rcfile') and args.rcfile is not None:
            read_locations = [str(args.rcfile)] + self._default_read_locations
        else:
            read_locations = self._default_read_locations
        resolved_locations = []
        for read_location in read_locations:
            resolved_locations.append(str(pathlib.Path(read_location).expanduser()))
        read_from = self._configparser.read(resolved_locations)
        self._logger.debug('Configuration read from {}'.format(str(read_from)))

    def __getitem__(self, key: str) -> typing.Any:
        namespaced_key = key.split('_')
        # Try once with nanaimo prefix (more specific)
        for x in range(len(namespaced_key)-1, 0, -1):
            try:
                group = '_'.join(namespaced_key[:x])
                value_key = '_'.join(namespaced_key[x:])
                return self._configparser['nanaimo:' + group][value_key]
            except KeyError:
                pass

        # Try again without the prefix
        for x in range(len(namespaced_key)-1, 0, -1):
            try:
                group = '_'.join(namespaced_key[:x])
                value_key = '_'.join(namespaced_key[x:])
                return self._configparser[group][value_key]
            except KeyError:
                pass

        return self._configparser['nanaimo'][key]

    def __contains__(self, key: str) -> bool:
        try:
            _ = self[key]
            return True
        except KeyError:
            return False

    def populate_default(self,
                         parser: argparse.ArgumentParser,
                         inout_args: typing.Tuple,
                         inout_kwargs: typing.Dict) -> None:

        # First replace any default with one from configuration.
        try:
            derived_key = self._derive_key_from_args(inout_args)
            from_config = self[derived_key]
            if from_config is not None:
                if 'type' in inout_kwargs:
                    inout_kwargs['default'] = inout_kwargs['type'](from_config)
                else:
                    inout_kwargs['default'] = from_config
                self._logger.debug('Setting the default for %s to %s (type %s) from a config file.',
                                   derived_key,
                                   inout_kwargs['default'],
                                   type(inout_kwargs['default']))
        except (KeyError, ValueError):
            pass

        # If we do have an environment variable and we allow this to be used
        # as a default it should become the default.
        if 'enable_default_from_environ' in inout_kwargs:
            if inout_kwargs['enable_default_from_environ']:
                self._handle_enable_default_from_environ(parser, inout_args, inout_kwargs)
            del inout_kwargs['enable_default_from_environ']

        # Finally, delete required if we managed to find a default.
        if 'required' in inout_kwargs and 'default' in inout_kwargs:
            # Since we have a default from configuration we can remove the required flag.
            del inout_kwargs['required']

    @classmethod
    def _derive_key_from_args(cls, inout_args: typing.Tuple) -> str:
        if len(inout_args) == 0:
            raise ValueError('No positional arguments')

        if len(inout_args) > 1:
            longform = str(inout_args[0]) if str(inout_args[0]).startswith('--') else str(inout_args[1])
        else:
            longform = str(inout_args[0])

        if not longform.startswith('--'):
            raise ValueError('No long-form argument to derive.')
        return longform[2:].replace('-', '_')

    def _handle_enable_default_from_environ(self,
                                            parser: argparse.ArgumentParser,
                                            inout_args: typing.Tuple,
                                            inout_kwargs: typing.Dict) -> None:
        name = self._derive_key_from_args(inout_args)

        env_variable = 'NANAIMO_{}'.format(name.upper().replace('-', '_'))
        try:
            parser_map = self._env_variable_index[parser]
            if env_variable in parser_map:
                raise RuntimeError('{} (derived from {}) was already derived from {}!'
                                   .format(env_variable, name,
                                           parser_map[env_variable]))
            parser_map[parser][env_variable] = name
        except KeyError:
            self._env_variable_index[parser] = {env_variable: name}
        self._set_default_from_environment(env_variable, inout_args, inout_kwargs)

    @classmethod
    def _set_default_from_environment(cls, env_variable: str, args: typing.Tuple, inout_kwargs: typing.Dict) -> None:
        if env_variable in os.environ:
            inout_kwargs['default'] = os.environ[env_variable]
            try:
                inout_kwargs['default'] = inout_kwargs['type'](inout_kwargs['default'])
            except KeyError:
                pass
            additional_help = 'Default value {} obtained from environment variable {}.'\
                .format(os.environ[env_variable], env_variable)
        else:
            additional_help = 'Set {} in the environment to override default.'.format(env_variable)
        try:
            inout_kwargs['help'] = inout_kwargs['help'] + '\n' + additional_help
        except KeyError:
            inout_kwargs['help'] = additional_help



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
#                                       (@@@@%%%%%%%%%&@@&.
#                              /%&&%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@@(
#                              *@&%%%%%%%%%&&%%%%%%%%%%%%%%%%%%&&&%%%%%%%
#                               @   @@@(@@@@%%%%%%%%%%%%%%%%&@@&* @@@   .
#                               ,   .        .  .@@@&                   /
#                                .       .                              *
#                               @@              .                       @
#                              @&&&&&&@. .    .                     *@%&@
#                              &&&&&&&&&&&&&&&&@@        *@@############@
#                     *&/ @@ #&&&&&&&&&&&&&&&&&&&&@  ###################*
#                              @&&&&&&&&&&&&&&&&&&##################@
#                                 %@&&&&&&&&&&&&&&################@
#                                        @&&&&&&&&&&%#######&@%
#  nanaimo                                   (@&&&&####@@*
#
import typing
import pytest

import nanaimo
import nanaimo.fixtures
import nanaimo.pytest.plugin


class Fixture(nanaimo.fixtures.SubprocessFixture):
    """
    Fixture base type that accepts a string argument ``cmd`` and executes it as a subprocess.

    .. invisible-code-block: python

        from nanaimo.fixtures import FixtureManager
        from nanaimo import Namespace, Arguments, Artifacts
        from nanaimo.builtin import nanaimo_cmd

        import os
        import asyncio

        loop = asyncio.get_event_loop()
        manager = FixtureManager(loop=loop)

    .. code-block:: python

        async def list_directory() -> Artifacts:

            cmd = nanaimo_cmd.Fixture(manager)

            return await cmd.gather(
                cmd_shell = ('dir' if os.name == 'nt' else 'ls -la')
            )

    .. invisible-code-block: python

        results = loop.run_until_complete(list_directory())

        assert 0 == results.result_code

    """

    fixture_name = 'nanaimo_cmd'
    argument_prefix = 'cmd'

    @classmethod
    def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
        arguments.add_argument('--cmd-shell', '-C', help='A shell command to run as a subprocess.')

    # +-----------------------------------------------------------------------+
    # | SubprocessFixture
    # +-----------------------------------------------------------------------+
    def on_construct_command(self, arguments: nanaimo.Namespace, inout_artifacts: nanaimo.Artifacts) -> str:
        """
        Called by the subprocess fixture to ask the specialization to form a command
        given a set of arguments.

        :param arguments: The arguments passed into :meth:`Fixture.on_gather`.
        :type arguments: nanaimo.Arguments
        :param inout_artifacts: A set of artifacts the superclass is assembling This
            is provided to the subclass to allow it to optionally contribute artifacts.
        :type inout_artifacts: nanaimo.Artifacts
        :return: The command to run in a subprocess shell.
        """
        optional_command = arguments.cmd_shell

        if isinstance(optional_command, str):
            return optional_command
        else:
            raise ValueError('No command provided (--cmd-shell, -C)')


@nanaimo.fixtures.PluggyFixtureManager.type_factory
def get_fixture_type() -> typing.Type['nanaimo.fixtures.Fixture']:
    return Fixture


@pytest.fixture
def nanaimo_cmd(request: typing.Any) -> nanaimo.fixtures.Fixture:
    """
    Provides a :class:`nanaimo.builtin.nanaimo_cmd.Fixture` fixture to a pytest.

    :param pytest_request: The request object passed into the pytest fixture factory.
    :type pytest_request: _pytest.fixtures.FixtureRequest
    :rtype: nanaimo.builtin.nanaimo_cmd.Fixture
    """
    return nanaimo.pytest.plugin.create_pytest_fixture(request, Fixture.get_canonical_name())



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#

import argparse
import asyncio
import contextlib
import pathlib
import typing
from unittest.mock import patch

import pytest

import nanaimo
import nanaimo.fixtures
from nanaimo.connections.uart import ConcurrentUart
from nanaimo.instruments.bkprecision import Series1900BUart


def create_dummy_serial_port_factory(fake_lines: typing.List[str],
                                     serial_simulator_type: typing.Type,
                                     event_loop: typing.Optional[asyncio.AbstractEventLoop] = None,
                                     loop_fake_data: bool = False) \
        -> Series1900BUart.UartFactoryType:

    @contextlib.contextmanager
    def dummy_serial_port(port: typing.Union[str, pathlib.Path]) -> typing.Generator:
        nonlocal fake_lines, event_loop
        with ConcurrentUart(serial_simulator_type(fake_lines, '\r'), event_loop, '\r', loop_fake_data) as bk_uart:
            print('''Yielding context with a simulated serial port set to emit {} line(s) of
fake output.'''.format(len(fake_lines)))
            yield bk_uart

    return dummy_serial_port


def to_namespace(fixture_type: typing.Type[nanaimo.fixtures.Fixture],
                 command: str,
                 nanaimo_defaults: nanaimo.config.ArgumentDefaults) -> nanaimo.Namespace:
    parser = argparse.ArgumentParser()
    fixture_type.on_visit_test_arguments(nanaimo.Arguments(parser, nanaimo_defaults))
    return nanaimo.Namespace(parser.parse_args(args=['--bk-port', 'dummy_port', '--bk-command', command]))


@pytest.mark.asyncio
@patch('nanaimo.fixtures.FixtureManager')
async def test_turn_off(MockFixtureManager: typing.Any,
                        paths_for_test: typing.Any,
                        event_loop: asyncio.AbstractEventLoop,
                        nanaimo_defaults: nanaimo.config.ArgumentDefaults,
                        serial_simulator_type: typing.Type) -> None:
    dummy_serial_port_factory = create_dummy_serial_port_factory(['OK'], serial_simulator_type, event_loop)
    args = to_namespace(Series1900BUart, '0', nanaimo_defaults)
    bk = Series1900BUart(nanaimo.fixtures.FixtureManager(),
                         args,
                         loop=event_loop,
                         uart_factory=dummy_serial_port_factory)
    assert 0 == int(await bk.gather())


@pytest.mark.asyncio
@patch('nanaimo.fixtures.FixtureManager')
async def test_turn_on(MockFixtureManager: typing.Any,
                       paths_for_test: typing.Any,
                       event_loop: asyncio.AbstractEventLoop,
                       nanaimo_defaults: nanaimo.config.ArgumentDefaults,
                       serial_simulator_type: typing.Type) -> None:
    dummy_serial_port_factory = create_dummy_serial_port_factory(['OK'], serial_simulator_type, event_loop)
    args = to_namespace(Series1900BUart, '1', nanaimo_defaults)
    bk = Series1900BUart(nanaimo.fixtures.FixtureManager(),
                         args,
                         loop=event_loop,
                         uart_factory=dummy_serial_port_factory)
    assert 0 == int(await bk.gather())


@pytest.mark.asyncio
@pytest.mark.timeout(10)
@patch('nanaimo.fixtures.FixtureManager')
async def test_turn_on_timeout(MockFixtureManager: typing.Any,
                               paths_for_test: typing.Any,
                               event_loop: asyncio.AbstractEventLoop,
                               nanaimo_defaults: nanaimo.config.ArgumentDefaults,
                               serial_simulator_type: typing.Type) -> None:
    args = to_namespace(Series1900BUart, '1', nanaimo_defaults)
    dummy_serial_port_factory = create_dummy_serial_port_factory(['NOPE'],
                                                                 serial_simulator_type,
                                                                 event_loop,
                                                                 loop_fake_data=False)
    bk = Series1900BUart(nanaimo.fixtures.FixtureManager(),
                         args,
                         loop=event_loop,
                         uart_factory=dummy_serial_port_factory)
    with pytest.raises(asyncio.TimeoutError):
        await bk.gather()


@pytest.mark.asyncio
@pytest.mark.timeout(10)
@patch('nanaimo.fixtures.FixtureManager')
async def test_turn_on_wait_for_voltage(MockFixtureManager: typing.Any,
                                        paths_for_test: typing.Any,
                                        event_loop: asyncio.AbstractEventLoop,
                                        nanaimo_defaults: nanaimo.config.ArgumentDefaults,
                                        serial_simulator_type: typing.Type) -> None:
    args = to_namespace(Series1900BUart, '1', nanaimo_defaults)
    mock_session = ['OK',
                    '000000000', 'OK',
                    '000100000', 'OK',
                    '001000000', 'OK',
                    '010000000', 'OK',
                    '010000000', 'OK',
                    '010000000', 'ERROR']
    dummy_serial_port_factory = create_dummy_serial_port_factory(mock_session,
                                                                 serial_simulator_type,
                                                                 event_loop,
                                                                 loop_fake_data=False)
    bk = Series1900BUart(nanaimo.fixtures.FixtureManager(),
                         args,
                         loop=event_loop,
                         uart_factory=dummy_serial_port_factory)
    assert 0 == int(await bk.gather(bk_target_voltage=1))


@pytest.mark.asyncio
@pytest.mark.timeout(10)
@patch('nanaimo.fixtures.FixtureManager')
async def test_turn_on_wait_for_voltage_timeout(MockFixtureManager: typing.Any,
                                                paths_for_test: typing.Any,
                                                event_loop: asyncio.AbstractEventLoop,
                                                nanaimo_defaults: nanaimo.config.ArgumentDefaults,
                                                serial_simulator_type: typing.Type) -> None:
    args = to_namespace(Series1900BUart, '1', nanaimo_defaults)
    mock_session = ['OK',
                    '000000000', 'OK',
                    '090000000', 'OK']
    dummy_serial_port_factory = create_dummy_serial_port_factory(mock_session,
                                                                 serial_simulator_type,
                                                                 event_loop,
                                                                 loop_fake_data=True)
    bk = Series1900BUart(nanaimo.fixtures.FixtureManager(),
                         args,
                         loop=event_loop,
                         uart_factory=dummy_serial_port_factory)
    with pytest.raises(asyncio.TimeoutError):
        await bk.gather(bk_target_voltage=10, bk_target_voltage_threshold_rising=.5)


@pytest.mark.asyncio
@patch('nanaimo.fixtures.FixtureManager')
async def test_get_display(MockFixtureManager: typing.Any,
                           paths_for_test: typing.Any,
                           event_loop: asyncio.AbstractEventLoop,
                           nanaimo_defaults: nanaimo.config.ArgumentDefaults,
                           serial_simulator_type: typing.Type) -> None:
    dummy_serial_port_factory = create_dummy_serial_port_factory(['030201451', 'OK'], serial_simulator_type, event_loop)
    args = to_namespace(Series1900BUart, '?', nanaimo_defaults)
    bk = Series1900BUart(nanaimo.fixtures.FixtureManager(),
                         args,
                         loop=event_loop,
                         uart_factory=dummy_serial_port_factory)
    artifacts = await bk.gather()
    display = artifacts.display
    assert 3.02 == display[0]
    assert 1.45 == display[1]
    assert Series1900BUart.ModeCC == display[2]



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
#                                       (@@@@%%%%%%%%%&@@&.
#                              /%&&%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@@(
#                              *@&%%%%%%%%%&&%%%%%%%%%%%%%%%%%%&&&%%%%%%%
#                               @   @@@(@@@@%%%%%%%%%%%%%%%%&@@&* @@@   .
#                               ,   .        .  .@@@&                   /
#                                .       .                              *
#                               @@              .                       @
#                              @&&&&&&@. .    .                     *@%&@
#                              &&&&&&&&&&&&&&&&@@        *@@############@
#                     *&/ @@ #&&&&&&&&&&&&&&&&&&&&@  ###################*
#                              @&&&&&&&&&&&&&&&&&&##################@
#                                 %@&&&&&&&&&&&&&&################@
#                                        @&&&&&&&&&&%#######&@%
#  nanaimo                                   (@&&&&####@@*
#
import asyncio
import re
import typing

import pytest

import nanaimo
import nanaimo.connections
import nanaimo.connections.uart
import nanaimo.fixtures
import nanaimo.pytest.plugin


class Fixture(nanaimo.fixtures.Fixture):
    """
    Gathers a log over a serial connection until a given pattern is matched.
    """

    fixture_name = 'nanaimo_serial_watch'
    argument_prefix = 'lw'

    def __init__(self,
                 manager: nanaimo.fixtures.FixtureManager,
                 args: nanaimo.Namespace,
                 **kwargs: typing.Any) -> None:
        super().__init__(manager, args, **kwargs)
        if 'uart_factory' in kwargs:
            self._uart_factory = typing.cast(typing.Callable[[str, int], nanaimo.connections.uart.ConcurrentUart],
                                             kwargs['uart_factory'])
        else:
            self._uart_factory = nanaimo.connections.uart.ConcurrentUart.new_default

    @classmethod
    def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
        nanaimo.connections.uart.ConcurrentUart.on_visit_test_arguments(arguments)
        arguments.add_argument('--lw-pattern',
                               default='.*',
                               enable_default_from_environ=True,
                               help='A python regular expression to search for')
        arguments.add_argument('--lw-disruption', default='\r\n')
        arguments.add_argument('--lw-disturb-rate', type=float)
        arguments.add_argument('--lw-update-period', type=float, default=1.0, help='Fractional time in seconds')

    async def on_gather(self, args: nanaimo.Namespace) -> nanaimo.Artifacts:
        """
        Watch the logs until the pattern matches.

        +--------------+---------------------------+-----------------------------------------------+
        | **Returned Artifacts**                                                                   |
        +--------------+---------------------------+-----------------------------------------------+
        | key          | type                      | Notes                                         |
        +==============+===========================+===============================================+
        | match        | re.MatchObject            | The match if result_code is 0                 |
        +--------------+---------------------------+-----------------------------------------------+
        | matched_line | str                       | The full line matched if result_code is 0     |
        +--------------+---------------------------+-----------------------------------------------+
        """
        with self._uart_factory(args.lw_port, args.lw_port_speed) as monitor:

            match_future, _ = await self.gate_tasks(self._matcher(args, monitor),
                                                    0.0,
                                                    self._agitator(args, monitor),
                                                    self._updater(args))

        artifacts = typing.cast(nanaimo.Artifacts, match_future.result())
        self._logger.info('Found match : %s', artifacts.matched_line)
        return artifacts

    # +-----------------------------------------------------------------------+
    # | PRIVATE
    # +-----------------------------------------------------------------------+
    async def _updater(self, args: nanaimo.Namespace) -> nanaimo.Artifacts:
        if args.lw_update_period is not None:
            while True:
                await asyncio.sleep(args.lw_update_period)
                self._logger.info('still waiting...')

        return nanaimo.Artifacts()

    async def _matcher(self, args: nanaimo.Namespace,
                       monitor: nanaimo.connections.uart.ConcurrentUart) -> nanaimo.Artifacts:
        artifacts = nanaimo.Artifacts()
        pattern = re.compile(args.lw_pattern)

        self._logger.debug('Starting to look for pattern : %s', str(pattern))
        while True:
            result = await monitor.get_line()
            match = pattern.search(result)
            if match:
                setattr(artifacts, 'match', match)
                setattr(artifacts, 'matched_line', result)
                break

        return artifacts

    async def _agitator(self, args: nanaimo.Namespace,
                        monitor: nanaimo.connections.uart.ConcurrentUart) -> nanaimo.Artifacts:
        if args.lw_disturb_rate is not None:
            while True:
                await asyncio.sleep(args.lw_disturb_rate)
                self._logger.debug('About to disturb the uart...')
                await monitor.put_line(args.lw_disruption)
        return nanaimo.Artifacts()


@nanaimo.fixtures.PluggyFixtureManager.type_factory
def get_fixture_type() -> typing.Type['Fixture']:
    return Fixture


@pytest.fixture
def nanaimo_serial_watch(request: typing.Any) -> nanaimo.fixtures.Fixture:
    """
    This Nanaimo fixture allows a test to monitor a serial port until a provided pattern matches.
    The `pytest fixture <https://docs.pytest.org/en/latest/fixture.html>`_ provides
    a :class:`nanaimo.builtin.nanaimo_serial_watch.Fixture` fixture to a pytest.

    :param pytest_request: The request object passed into the pytest fixture factory.
    :type pytest_request: _pytest.fixtures.FixtureRequest
    :rtype: nanaimo.builtin.nanaimo_serial_watch.Fixture
    """
    return nanaimo.pytest.plugin.create_pytest_fixture(request, Fixture.get_canonical_name())



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
#                                       (@@@@%%%%%%%%%&@@&.
#                              /%&&%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@@(
#                              *@&%%%%%%%%%&&%%%%%%%%%%%%%%%%%%&&&%%%%%%%
#                               @   @@@(@@@@%%%%%%%%%%%%%%%%&@@&* @@@   .
#                               ,   .        .  .@@@&                   /
#                                .       .                              *
#                               @@              .                       @
#                              @&&&&&&@. .    .                     *@%&@
#                              &&&&&&&&&&&&&&&&@@        *@@############@
#                     *&/ @@ #&&&&&&&&&&&&&&&&&&&&@  ###################*
#                              @&&&&&&&&&&&&&&&&&&##################@
#                                 %@&&&&&&&&&&&&&&################@
#                                        @&&&&&&&&&&%#######&@%
#  nanaimo                                   (@&&&&####@@*
#
import typing

import pytest

import nanaimo
import nanaimo.fixtures
import nanaimo.pytest.plugin


class Fixture(nanaimo.fixtures.SubprocessFixture):
    """
    This fixture assumes that ssh is available and functional on the system.
    """

    fixture_name = 'nanaimo_ssh'
    argument_prefix = 'ssh'

    @classmethod
    def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
        arguments.add_argument('--ssh-target',
                               help='The IP or hostname for the target system.')
        arguments.add_argument('--ssh-as-user',
                               help='The user to upload as.')
        arguments.add_argument('--ssh-command',
                               help='The command to run.')
        arguments.add_argument('--ssh-identity',
                               help='The identify file to use')

    def on_construct_command(self, args: nanaimo.Namespace, inout_artifacts: nanaimo.Artifacts) -> str:
        cmd = 'ssh {port} {ident} {user}@{target} \'{command}\''.format(
            port='-P {}'.format(args.ssh_port) if args.ssh_port is not None else '',
            command=str(args.ssh_command),
            user=args.ssh_as_user,
            target=args.ssh_target,
            ident='-i {}'.format(args.ssh_identity) if args.ssh_identity is not None else ''
        )
        return cmd


@nanaimo.fixtures.PluggyFixtureManager.type_factory
def get_fixture_type() -> typing.Type['Fixture']:
    return Fixture


@pytest.fixture
def nanaimo_ssh(request: typing.Any) -> nanaimo.fixtures.Fixture:
    """
    The SSH fixture allows a test to run arbitrary commands on a remote device.
    This `pytest fixture <https://docs.pytest.org/en/latest/fixture.html>`_ provides
    a :class:`nanaimo.builtin.nanaimo_ssh.Fixture` fixture to a pytest.

    :param pytest_request: The request object passed into the pytest fixture factory.
    :type pytest_request: _pytest.fixtures.FixtureRequest
    :rtype: nanaimo.builtin.nanaimo_ssh.Fixture
    """
    return nanaimo.pytest.plugin.create_pytest_fixture(request, Fixture.get_canonical_name())



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
#                                       (@@@@%%%%%%%%%&@@&.
#                              /%&&%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@@(
#                              *@&%%%%%%%%%&&%%%%%%%%%%%%%%%%%%&&&%%%%%%%
#                               @   @@@(@@@@%%%%%%%%%%%%%%%%&@@&* @@@   .
#                               ,   .        .  .@@@&                   /
#                                .       .                              *
#                               @@              .                       @
#                              @&&&&&&@. .    .                     *@%&@
#                              &&&&&&&&&&&&&&&&@@        *@@############@
#                     *&/ @@ #&&&&&&&&&&&&&&&&&&&&@  ###################*
#                              @&&&&&&&&&&&&&&&&&&##################@
#                                 %@&&&&&&&&&&&&&&################@
#                                        @&&&&&&&&&&%#######&@%
#  nanaimo                                   (@&&&&####@@*
#
"""
This module contains control and communication objects for popular instruments
useful to the type of testing Nanaimo is focused on. Where possible these async
interfaces will use pure-python to communicate directly with an instrument
using known protocols and well supported communication busses (e.g. ethernet, uart, or CAN),
however, some instruments will require the installation of a vendor
CLI for the Nanaimo automation to use.

.. Warning::
    Do not merge third-party code into the Nanaimo repo or provide Nanaimo abstractions
    that are incompatible with the licensing of this software.

.. _jlink-example:

.. figure:: https://thirtytwobits.github.io/nanaimo/images/jlink_example.png
   :alt: Example block diagram of a Nanaimo instrument using a vendor CLI.

   Example of a Nanaimo Segger JLink instrument using the JLinkExe CLI as a Python
   subprocess.

:numref:`jlink-example` shows the Nanaimo jlink instrument using `Segger's JLink Commander`_
as a subprocess to upload a firmware image to a microcontroller.

.. _`Segger's JLink Commander`: https://www.segger.com/products/debug-probes/j-link/tools/j-link-commander/

"""



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
import argparse
import asyncio
import pathlib

import pytest

import nanaimo
import nanaimo.builtin
import nanaimo.fixtures
from nanaimo.builtin import nanaimo_gather


class CanonicallyNamed(nanaimo.fixtures.Fixture):

    fixture_name = 'foo_bar'


def test_canonical_name(dummy_nanaimo_fixture: nanaimo.fixtures.Fixture) -> None:
    """
    Verify the behavior of Fixture.get_canonical_name()
    """
    assert 'material.DummyFixture' == type(dummy_nanaimo_fixture).get_canonical_name()
    assert CanonicallyNamed.fixture_name == CanonicallyNamed.get_canonical_name()


@pytest.mark.asyncio
async def test_gather_coroutines(nanaimo_fixture_manager: nanaimo.fixtures.FixtureManager) -> None:

    parser = argparse.ArgumentParser()
    nanaimo_gather.Fixture.on_visit_test_arguments(nanaimo.Arguments(parser))
    args = nanaimo.Namespace(parser.parse_args(['--gather-coroutine', 'nanaimo_bar',
                                                '--gather-coroutine', 'nanaimo_bar']))

    gather_fixture = nanaimo_gather.Fixture(nanaimo_fixture_manager, args)

    results = await gather_fixture.gather()

    assert results.result_code == 0


@pytest.mark.timeout(10)
@pytest.mark.asyncio
async def test_observe_tasks(dummy_nanaimo_fixture: nanaimo.fixtures.Fixture) -> None:
    """
    Test the observe_tasks method of Fixture
    """

    async def evaluating() -> int:
        return 0

    async def running() -> int:
        waits = 2
        while waits > 0:
            await asyncio.sleep(.1)
            waits -= 1
        return 1

    result = await dummy_nanaimo_fixture.observe_tasks_assert_not_done(evaluating(),
                                                                       0,
                                                                       running())
    assert len(result) == 1
    should_be_running = result.pop()

    assert not should_be_running.done()

    assert 1 == await should_be_running


@pytest.mark.timeout(10)
@pytest.mark.asyncio
async def test_observe_tasks_failure(dummy_nanaimo_fixture: nanaimo.fixtures.Fixture) -> None:
    """
    Test the observe_tasks method of Fixture where the running tasks exit.
    """

    async def evaluating() -> int:
        waits = 2
        while waits > 0:
            await asyncio.sleep(.1)
            waits -= 1
        return 1

    async def running() -> int:
        return 1

    with pytest.raises(nanaimo.AssertionError):
        await dummy_nanaimo_fixture.observe_tasks_assert_not_done(evaluating(),
                                                                  0,
                                                                  running())


@pytest.mark.timeout(10)
@pytest.mark.asyncio
async def test_observe_tasks_failure_no_assert(dummy_nanaimo_fixture: nanaimo.fixtures.Fixture) -> None:
    """
    Test the observe_tasks method of Fixture where the running tasks exit but without throwing
    an assertion error.
    """

    async def evaluating() -> int:
        waits = 2
        while waits > 0:
            await asyncio.sleep(.1)
            waits -= 1
        return 1

    async def running() -> int:
        return 1

    result = await dummy_nanaimo_fixture.observe_tasks(evaluating(),
                                                       0,
                                                       running())

    assert 0 == len(result)


@pytest.mark.timeout(10)
@pytest.mark.asyncio
async def test_observe_tasks_timeout(dummy_nanaimo_fixture: nanaimo.fixtures.Fixture) -> None:
    """
    Test the observe_tasks method of Fixture where the running tasks do not exit.
    """

    async def evaluating() -> int:
        while True:
            await asyncio.sleep(1)

    async def running() -> int:
        while True:
            await asyncio.sleep(1)

    with pytest.raises(asyncio.TimeoutError):
        await dummy_nanaimo_fixture.observe_tasks_assert_not_done(evaluating(),
                                                                  1,
                                                                  running())


@pytest.mark.timeout(20)
@pytest.mark.asyncio
async def test_countdown_sleep(dummy_nanaimo_fixture: nanaimo.fixtures.Fixture) -> None:
    """
    Test the observe_tasks method of Fixture where the running tasks do not exit.
    """
    await dummy_nanaimo_fixture.countdown_sleep(5.3)


@pytest.mark.timeout(20)
@pytest.mark.asyncio
async def test_gather_timeout(gather_timeout_fixture: nanaimo.fixtures.Fixture) -> None:
    """
    Test the standard fixture timeout.
    """
    gather_timeout_fixture.gather_timeout_seconds = 1.0
    with pytest.raises(asyncio.TimeoutError):
        await gather_timeout_fixture.gather()


@pytest.mark.asyncio
async def test_subprocess_fixture() -> None:
    """
    Verify the subprocess fixture base class.
    """
    import nanaimo.version

    class SubprocessTestHarness(nanaimo.fixtures.SubprocessFixture):

        @classmethod
        def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
            pass

        def on_construct_command(self, arguments: nanaimo.Namespace, inout_artifacts: nanaimo.Artifacts) -> str:
            setattr(inout_artifacts, 'foo', 'bar')
            return 'nait --version'

    subject = SubprocessTestHarness(nanaimo.fixtures.FixtureManager())
    artifacts = await subject.gather()

    assert 'bar' == artifacts.foo
    assert artifacts.stdout == nanaimo.version.__version__


@pytest.mark.asyncio
async def test_subprocess_fixture_logfile(build_output: pathlib.Path) -> None:
    """
    Verify the subprocess fixture base class.
    """
    import nanaimo.version

    logfile = build_output / pathlib.Path('test_subprocess_fixture_logfile').with_suffix('.log')
    class SubprocessTestHarness(nanaimo.fixtures.SubprocessFixture):

        @classmethod
        def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
            pass

        def on_construct_command(self, arguments: nanaimo.Namespace, inout_artifacts: nanaimo.Artifacts) -> str:

            setattr(inout_artifacts, 'logfile', str(logfile))
            return 'nait --version'

    subject = SubprocessTestHarness(nanaimo.fixtures.FixtureManager())
    artifacts = await subject.gather()

    assert artifacts.result_code == 0

    found = False
    with open(str(artifacts.logfile), 'r') as logfile_io:
        for line in logfile_io:
            print(line)
            if line.find(nanaimo.version.__version__) != -1:
                found = True
                break

    assert found



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
#                                       (@@@@%%%%%%%%%&@@&.
#                              /%&&%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@@(
#                              *@&%%%%%%%%%&&%%%%%%%%%%%%%%%%%%&&&%%%%%%%
#                               @   @@@(@@@@%%%%%%%%%%%%%%%%&@@&* @@@   .
#                               ,   .        .  .@@@&                   /
#                                .       .                              *
#                               @@              .                       @
#                              @&&&&&&@. .    .                     *@%&@
#                              &&&&&&&&&&&&&&&&@@        *@@############@
#                     *&/ @@ #&&&&&&&&&&&&&&&&&&&&@  ###################*
#                              @&&&&&&&&&&&&&&&&&&##################@
#                                 %@&&&&&&&&&&&&&&################@
#                                        @&&&&&&&&&&%#######&@%
#  nanaimo                                   (@&&&&####@@*
#
"""
Almost everything in Nanaimo is a :class:`nanaimo.fixtures.Fixture`. Fixtures can be pytest fixtures, instrument
abstractions, aggregates of other fixtures, or anything else that makes sense. The important thing
is that any fixture can be a pytest fixture or can be awaited directly using :ref:`nait`.
"""
import abc
import asyncio
import io
import logging
import math
import typing

import pluggy

import nanaimo


class Fixture(metaclass=abc.ABCMeta):
    """
    Common, abstract class for pytest fixtures based on Nanaimo. Nanaimo fixtures provide a visitor pattern for
    arguments that are common for both pytest extra arguments and for argparse commandline arguments. This allows
    a Nanaimo fixture to expose a direct invocation mode for debuging with the same arguments used by the fixture
    as a pytest plugin. Additionally all Nanaimo fixtures provide a :func:`gather` function that takes a
    :class:`nanaimo.Namespace` containing the provided arguments and returns a set of :class:`nanaimo.Artifacts`
    gathered by the fixture. The contents of these artifacts are documented by each concrete fixture.

    .. invisible-code-block: python

        import nanaimo
        import nanaimo.fixtures
        import asyncio

        _doc_loop = asyncio.new_event_loop()

    .. code-block:: python

        class MyFixture(nanaimo.fixtures.Fixture):
            @classmethod
            def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
                arguments.add_argument('--foo', default='bar')

            async def on_gather(self, args: nanaimo.Namespace) -> nanaimo.Artifacts:
                artifacts = nanaimo.Artifacts(-1)
                # do something and then return
                artifacts.result_code = 0
                return artifacts

    .. invisible-code-block: python

        foo = MyFixture(nanaimo.fixtures.FixtureManager(), nanaimo.Namespace(), loop=_doc_loop)

        _doc_loop.run_until_complete(foo.gather())

    `MyFixture` can now be used from a commandline like::

        python -m nanaimo MyFixture --foo baz

    or as part of a pytest::

        pytest --foo=baz

    :param FixtureManager manager: The fixture manager that is the scope for this fixture. There must be
        a 1:1 relationship between a fixture instance and a fixture manager instance.
    :param nanaimo.Namespace args: A namespace containing the arguments for this fixture.
    :param kwargs: All fixtures can be given a :class:`asyncio.AbstractEventLoop` instance to use as ``loop`` and
        an initial value for :data:`gather_timeout_seconds` as ``gather_timeout_seconds (float)``. Other keyword
        arguments may used by fixture specializations.
    """

    @classmethod
    def get_canonical_name(cls) -> str:
        """
        The name to use as a key for this :class:`nanaimo.fixtures.Fixture` type.
        If a class defines a string `fixture_name` this will be used
        as the canonical name otherwise it will be the name of the
        fixture class itself.

        .. invisible-code-block: python
            import nanaimo
            import nanaimo.fixtures

        .. code-block:: python

            class MyFixture(nanaimo.fixtures.Fixture):

                fixture_name = 'my_fixture'

            assert 'my_fixture' == MyFixture.get_canonical_name()

        """
        return str(getattr(cls, 'fixture_name', '.'.join([cls.__module__, cls.__qualname__])))

    @classmethod
    def get_argument_prefix(cls) -> str:
        """
        The name to use as a prefix for arguments. This also becomes the configuration
        section that the fixture's arguments can be overridden from. If the fixture
        defines an ``argument_prefix`` class member this value is used otherwise the
        value returned from :meth:`get_canonical_name` is used.

        .. invisible-code-block: python
            import nanaimo
            import nanaimo.fixtures

        .. code-block:: python

            class MyFixture(nanaimo.fixtures.Fixture):

                argument_prefix = 'mf'

        >>> MyFixture.get_argument_prefix()  # noqa : F821
        'mf'

        .. code-block:: python

            class MyOtherFixture(nanaimo.fixtures.Fixture):
                # this class doesn't define argument_prefix so
                # so the canonical name is used instead.
                fixture_name = 'my_outre_fixture'

        >>> MyOtherFixture.get_argument_prefix()  # noqa : F821
        'my-outre-fixture'

        """
        return str(getattr(cls, 'argument_prefix', cls.get_canonical_name().replace('_', '-')))

    def __init__(self,
                 manager: 'FixtureManager',
                 args: typing.Optional[nanaimo.Namespace] = None,
                 **kwargs: typing.Any):
        self._manager = manager
        self._args = (args if args is not None else nanaimo.Namespace())
        self._name = self.get_canonical_name()
        self._logger = logging.getLogger(self._name)
        if 'loop' in kwargs:
            self._loop = typing.cast(typing.Optional[asyncio.AbstractEventLoop], kwargs['loop'])
        else:
            self._loop = None
        if 'gather_timeout_seconds' in kwargs:
            gather_timeout_seconds = typing.cast(typing.Optional[float], kwargs['gather_timeout_seconds'])
            self._gather_timeout_seconds = (gather_timeout_seconds if gather_timeout_seconds is not None else 0)
        else:
            self._gather_timeout_seconds = 0

    async def gather(self, *args: typing.Any, **kwargs: typing.Any) -> nanaimo.Artifacts:
        """
        Coroutine awaited to gather a new set of fixture artifacts.

        :param kwargs: Optional arguments to override or augment the arguments provided to the
            :class:`nanaimo.fixtures.Fixture` constructor
        :return: A set of artifacts with the :attr:`nanaimo.Artifacts.result_code` set to indicate the success or
                 failure of the fixture's artifact gathering activies.
        :raises asyncio.TimeoutError: If :data:`gather_timeout_seconds` is > 0 and :meth:`on_gather` takes longer
            then this to complete or if on_gather itself raises a timeout error.
        """
        pushed_args = self._args
        self._args = self._args.merge(**kwargs)
        try:
            routine = self.on_gather(self._args)
            if self._gather_timeout_seconds > 0:
                done, pending = await asyncio.wait([asyncio.ensure_future(routine)],
                                                   loop=self.manager.loop,
                                                   timeout=self._gather_timeout_seconds,
                                                   return_when=asyncio.ALL_COMPLETED)
                if len(pending) > 0:
                    pending.pop().cancel()
                    raise asyncio.TimeoutError('{} gather was cancelled after waiting for {} seconds'
                                               .format(self.get_canonical_name(),
                                                       self._gather_timeout_seconds))
                return done.pop().result()
            else:
                return await routine
        finally:
            self._args = pushed_args

    # +-----------------------------------------------------------------------+
    # | PROPERTIES
    # +-----------------------------------------------------------------------+
    @property
    def name(self) -> str:
        """
        The canonical name for the Fixture.
        """
        return self._name

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        """
        The running asyncio EventLoop in use by this Fixture.
        This will be the loop provided to the fixture in the constructor if that loop is still
        running otherwise the loop will be a running loop retrieved by :func:`asyncio.get_event_loop`.
        :raises RuntimeError: if no running event loop could be found.
        """
        if self._loop is None or not self._loop.is_running():
            self._loop = self.manager.loop
        if not self._loop.is_running():
            raise RuntimeError('No running event loop was found!')
        return self._loop

    @property
    def manager(self) -> 'FixtureManager':
        """
        The :class:`FixtureManager` that owns this :class:`nanaimo.fixtures.Fixture`.
        """
        return self._manager

    @property
    def logger(self) -> logging.Logger:
        """
        A logger for this :class:`nanaimo.fixtures.Fixture` instance.
        """
        return self._logger

    @property
    def fixture_arguments(self) -> nanaimo.Namespace:
        """
        The Fixture-wide arguments. Can be overridden by kwargs for each :meth:`gather` invocation.
        """
        return self._args

    @property
    def gather_timeout_seconds(self) -> float:
        """
        The timeout in fractional seconds to wait for :meth:`on_gather` to complete before raising
        a :class:`asyncio.TimeoutError`.
        """
        return self._gather_timeout_seconds

    @gather_timeout_seconds.setter
    def gather_timeout_seconds(self, gather_timeout_seconds: float) -> None:
        self._gather_timeout_seconds = gather_timeout_seconds

    # +-----------------------------------------------------------------------+
    # | ABSTRACT METHODS
    # +-----------------------------------------------------------------------+
    @classmethod
    @abc.abstractmethod
    def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
        """
        Called by the environment before instantiating any :class:`nanaimo.fixtures.Fixture` instances to register
        arguments supported by each type. These arguments should be portable between both :mod:`argparse`
        and ``pytest``. The fixture is registered for this callback by returning a reference to its
        type from a :attr:`PluggyFixtureManager.type_factory` annotated function registered as an
        entrypoint in the Python application.
        """
        ...

    @abc.abstractmethod
    async def on_gather(self, args: nanaimo.Namespace) -> nanaimo.Artifacts:
        """
        Coroutine awaited by a call to :meth:`gather`. The fixture should always retrieve new artifacts when invoked
        leaving caching to the caller.

        :param args: The arguments provided for the fixture instance merged with kwargs provided to the :meth:`gather`
            method.
        :type args: nanaimo.Namespace
        :return: A set of artifacts with the :attr:`nanaimo.Artifacts.result_code` set to indicate the success or
            failure of the fixture's artifact gathering activies.
        :raises asyncio.TimeoutError: It is valid for a fixture to raise timeout errors from this method.
        """
        ...

    # +-----------------------------------------------------------------------+
    # | ASYNC HELPERS
    # +-----------------------------------------------------------------------+

    async def countdown_sleep(self, sleep_time_seconds: float) -> None:
        """
        Calls :func:`asyncio.sleep` for 1 second then emits an :meth:`logging.Logger.info`
        of the time remaining until `sleep_time_seconds`.
        This is useful for long waits as an indication that the process is not deadlocked.

        :param sleep_time_seconds:  The amount of time in seconds for this coroutine to wait
            before exiting. For each second that passes while waiting for this amount of time
            the coroutine will :func:`asyncio.sleep` for 1 second.
        :type sleep_time_seconds: float
        """
        count_down = sleep_time_seconds
        while count_down >= 0:
            self._logger.info('%d', math.ceil(count_down))
            await asyncio.sleep(1)
            count_down -= 1

    async def observe_tasks_assert_not_done(self,
                                            observer_co_or_f: typing.Union[typing.Coroutine, asyncio.Future],
                                            timeout_seconds: float,
                                            *persistent_tasks: typing.Union[typing.Coroutine, asyncio.Future]) \
            -> typing.Set[asyncio.Future]:
        """
        Allows running a set of tasks but returning when an observer task completes. This allows a pattern where
        a single task is evaluating the side-effects of other tasks as a gate to continuing the test.

        :param observer_co_or_f: The task that is expected to complete in less than ``timeout_seconds``.
        :type observer_co_or_f: typing.Union[typing.Coroutine, asyncio.Future]
        :param float timeout_seconds: Time in seconds to observe for before raising :class:`asyncio.TimeoutError`.
            Set to 0 to disable.
        :param persistent_tasks: Iterable of tasks that must remain active or :class:`AssertionError` will be raised.
        :type persistent_tasks: typing.Union[typing.Coroutine, asyncio.Future]

        :return: a list of the persistent tasks as futures.
        :rtype: typing.Set[asyncio.Future]

        :raises AssertionError: if any of the persistent tasks exited.
        :raises asyncio.TimeoutError: if the observer task does not complete within ``timeout_seconds``.
        """
        _, _, done, pending = await self._observe_tasks(observer_co_or_f, timeout_seconds, False, *persistent_tasks)
        if len(done) > 1:
            raise nanaimo.AssertionError('Tasks under observation completed before the observation was complete.')
        return pending

    async def observe_tasks(self,
                            observer_co_or_f: typing.Union[typing.Coroutine, asyncio.Future],
                            timeout_seconds: float,
                            *persistent_tasks: typing.Union[typing.Coroutine, asyncio.Future]) \
            -> typing.Set[asyncio.Future]:
        """
        Allows running a set of tasks but returning when an observer task completes. This allows a pattern where
        a single task is evaluating the side-effects of other tasks as a gate to continuing the test or simply
        that a set of task should continue to run but a single task must complete.

        :param observer_co_or_f: The task that is expected to complete in less than ``timeout_seconds``.
        :type observer_co_or_f: typing.Union[typing.Coroutine, asyncio.Future]
        :param float timeout_seconds: Time in seconds to observe for before raising :class:`asyncio.TimeoutError`.
            Set to 0 to disable.
        :param persistent_tasks: Iterable of tasks that may remain active.
        :type persistent_tasks: typing.Union[typing.Coroutine, asyncio.Future]

        :return: a list of the persistent tasks as futures.
        :rtype: typing.Set[asyncio.Future]

        :raises AssertionError: if any of the persistent tasks exited.
        :raises asyncio.TimeoutError: if the observer task does not complete within ``timeout_seconds``.
        """

        _, _, done, pending = await self._observe_tasks(observer_co_or_f, timeout_seconds, False, *persistent_tasks)
        return pending

    async def gate_tasks(self,
                         gate_co_or_f: typing.Union[typing.Coroutine, asyncio.Future],
                         timeout_seconds: float,
                         *gated_tasks: typing.Union[typing.Coroutine, asyncio.Future]) \
            -> typing.Tuple[asyncio.Future, typing.List[asyncio.Future]]:
        """
        Runs a set of tasks until a gate task completes then cancels the remaining tasks.

        .. invisible-code-block: python

            from nanaimo.fixtures import FixtureManager
            from nanaimo.builtin import nanaimo_bar
            import asyncio

            loop = asyncio.get_event_loop()
            manager = FixtureManager(loop=loop)

        .. code-block:: python

            async def gated_task():
                while True:
                    await asyncio.sleep(.1)

            async def gate_task():
                await asyncio.sleep(1)
                return 'gate passed'

            async def example():

                any_fixture = nanaimo_bar.Fixture(manager)

                gate_future, gated_futures = await any_fixture.gate_tasks(gate_task(), 0, gated_task())

                assert not gate_future.cancelled()
                assert 'gate passed' == gate_future.result()
                assert len(gated_futures) == 1
                assert gated_futures[0].cancelled()

        .. invisible-code-block: python

            loop.run_until_complete(example())

        :param gate_co_or_f: The task that is expected to complete in less than ``timeout_seconds``.
        :type gate_co_or_f: typing.Union[typing.Coroutine, asyncio.Future]
        :param float timeout_seconds: Time in seconds to wait for the gate for before raising
            :class:`asyncio.TimeoutError`. Set to 0 to disable.
        :param persistent_tasks: Iterable of tasks that may remain active.
        :type persistent_tasks: typing.Union[typing.Coroutine, asyncio.Future]

        :return: a tuple of the gate future and a set of the gated futures.
        :rtype: typing.Tuple[asyncio.Future, typing.Set[asyncio.Future]]:

        :raises AssertionError: if any of the persistent tasks exited.
        :raises asyncio.TimeoutError: if the observer task does not complete within ``timeout_seconds``.
        """

        observer, observed, _, _ = await self._observe_tasks(gate_co_or_f, timeout_seconds, True, *gated_tasks)
        return observer, observed

    # +-----------------------------------------------------------------------+
    # | PRIVATE
    # +-----------------------------------------------------------------------+
    @classmethod
    async def _do_cancel_remaining(cls, pending: typing.Set[asyncio.Future]) -> None:
        for f in pending:
            f.cancel()
            try:
                await f
            except asyncio.CancelledError:
                pass

    async def _observe_tasks(self,
                             observer_co_or_f: typing.Union[typing.Coroutine, asyncio.Future],
                             timeout_seconds: float,
                             cancel_remaining: bool,
                             *args: typing.Union[typing.Coroutine, asyncio.Future]) -> \
            typing.Tuple[asyncio.Future, typing.List[asyncio.Future],
                         typing.Set[asyncio.Future], typing.Set[asyncio.Future]]:
        """
        :returns: observer, observed, done, pending
        """
        did_timeout = False
        observer_future = asyncio.ensure_future(observer_co_or_f)
        the_children_are_our_futures = [observer_future]
        observed_futures = []
        done_done = set()  # type: typing.Set[asyncio.Future]
        for co_or_f in args:
            o = asyncio.ensure_future(co_or_f)
            the_children_are_our_futures.append(o)
            observed_futures.append(o)

        start_time = self.loop.time()
        wait_timeout = (timeout_seconds if timeout_seconds > 0 else None)

        while True:
            done, pending = await asyncio.wait(
                the_children_are_our_futures,
                timeout=wait_timeout,
                return_when=asyncio.FIRST_COMPLETED)

            for d in done:
                the_children_are_our_futures.remove(d)
                done_done.add(d)

            if observer_future.done():
                break

            if wait_timeout is not None and self.loop.time() - start_time > wait_timeout:
                did_timeout = True
                break

        if cancel_remaining:
            await self._do_cancel_remaining(pending)
            done_done.update(pending)
            pending.clear()

        if did_timeout:
            raise asyncio.TimeoutError()
        else:
            return observer_future, observed_futures, done_done, pending


# +---------------------------------------------------------------------------+


class SubprocessFixture(Fixture):
    """
    Fixture base type that accepts a string argument ``cmd`` and executes it as a subprocess.
    """

    async def on_gather(self, args: nanaimo.Namespace) -> nanaimo.Artifacts:
        """
        +--------------+---------------------------+-----------------------------------------------+
        | **Returned Artifacts**                                                                   |
        +--------------+---------------------------+-----------------------------------------------+
        | key          | type                      | Notes                                         |
        +==============+===========================+===============================================+
        | logfile      | Optional[str]             | A file to log to. Set this in                 |
        |              |                           | on_construct_command and this baseclass will  |
        |              |                           | multiplex all stdout and stderr to the file.  |
        +--------------+---------------------------+-----------------------------------------------+
        | stderr       | Optional[str]             | Capture of stderr                             |
        +--------------+---------------------------+-----------------------------------------------+
        | stdout       | Optional[str]             | Capture of stdout                             |
        +--------------+---------------------------+-----------------------------------------------+
        """
        artifacts = nanaimo.Artifacts()

        cmd = self.on_construct_command(args, artifacts)

        logfile_handler = None  # type: typing.Optional[logging.FileHandler]
        if artifacts.logfile is not None:
            logfile_handler = logging.FileHandler(filename=str(artifacts.logfile), mode='a')
            file_formatter = logging.Formatter(fmt='%(asctime)s %(levelname)s %(name)s: %(message)s',
                                               datefmt='%Y-%m-%d %H:%M:%S')
            logfile_handler.setFormatter(file_formatter)
            self._logger.addHandler(logfile_handler)

        try:
            self._logger.debug('About to execute command "%s" in a subprocess shell', cmd)

            proc = await asyncio.create_subprocess_shell(
                cmd,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE
            )  # type: asyncio.subprocess.Process

            stdout, stderr = await self._wait_for_either_until_neither(
                (proc.stdout if proc.stdout is not None else self._NoopStreamReader()),
                (proc.stderr if proc.stderr is not None else self._NoopStreamReader()))

            await proc.wait()

            self._logger.debug('command "%s" exited with %i', cmd, proc.returncode)

            setattr(artifacts, 'stdout', stdout)
            setattr(artifacts, 'stderr', stderr)

            artifacts.result_code = proc.returncode
            return artifacts
        finally:
            if logfile_handler is not None:
                self._logger.removeHandler(logfile_handler)

    # +-----------------------------------------------------------------------+
    # | ABSTRACT METHODS
    # +-----------------------------------------------------------------------+
    @abc.abstractmethod
    def on_construct_command(self, arguments: nanaimo.Namespace, inout_artifacts: nanaimo.Artifacts) -> str:
        """
        Called by the subprocess fixture to ask the specialization to form a command
        given a set of arguments.

        :param arguments: The arguments passed into :meth:`Fixture.on_gather`.
        :type arguments: nanaimo.Arguments
        :param inout_artifacts: A set of artifacts the superclass is assembling This
            is provided to the subclass to allow it to optionally contribute artifacts.
        :type inout_artifacts: nanaimo.Artifacts
        :return: The command to run in a subprocess shell.
        """
        ...

    # +-----------------------------------------------------------------------+
    # | PRIVATE METHODS
    # +-----------------------------------------------------------------------+

    class _NoopStreamReader(asyncio.StreamReader):

        def __init__(self) -> None:
            super().__init__()
            self.feed_eof()

    async def _wait_for_either_until_neither(self,
                                             stdout: asyncio.StreamReader,
                                             stderr: asyncio.StreamReader) \
            -> typing.Tuple[str, str]:
        """
        Wait for a line of data from either stdout or stderr and log this data as received.
        When both are EOF then exit.

        :returns: Tuple of stdout, stderr
        """
        stdout_buffer = io.StringIO()
        stderr_buffer = io.StringIO()

        future_out = asyncio.ensure_future(stdout.readline())
        future_err = asyncio.ensure_future(stderr.readline())

        pending = set([future_out, future_err])

        while len(pending) > 0:

            done, pending = await asyncio.wait(pending)

            for future_done in done:
                result = future_done.result().strip()
                if len(result) > 0:
                    line = result.decode()
                    if future_done == future_err:
                        future_err = asyncio.ensure_future(stderr.readline())
                        pending.add(future_err)
                        stderr_buffer.write(line)
                        self._logger.error(line)
                    else:
                        future_out = asyncio.ensure_future(stdout.readline())
                        pending.add(future_out)
                        stdout_buffer.write(line)
                        self._logger.info(line.strip())
        return (stdout_buffer.getvalue(), stderr_buffer.getvalue())


# +---------------------------------------------------------------------------+


class FixtureManager:
    """
    A simple fixture manager and a baseclass for specalized managers.
    """

    def __init__(self, loop: typing.Optional[asyncio.AbstractEventLoop] = None) -> None:
        self._loop = loop

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        """
        The running asyncio EventLoop in use by all Fixtures.
        This will be the loop provided to the fixture manager in the constructor if that loop is still
        running otherwise the loop will be a running loop retrieved by :func:`asyncio.get_event_loop`.
        :raises RuntimeError: if no running event loop could be found.
        """
        if self._loop is None or not self._loop.is_running():
            self._loop = asyncio.get_event_loop()
        if not self._loop.is_running():
            raise RuntimeError('No running event loop was found!')
        return self._loop

    def fixture_types(self) -> typing.Generator:
        """
        Yields each fixture type registered with this object.
        """
        if False:
            yield

    def create_fixture(self,
                       canonical_name: str,
                       args: typing.Optional[nanaimo.Namespace] = None,
                       loop: typing.Optional[asyncio.AbstractEventLoop] = None) -> Fixture:
        """
        Create a new :class:`nanaimo.fixtures.Fixture` instance iff the ``canonical_name``` is a registered
        plugin for this process.

        :param str canonical_name: The canonical name of the fixture to instantiate.
        :param nanaimo.Namespace args: The arguments to provide to the new instance.
        :param loop: An event loop to provide the fixture instance.
        :type loop: typing.Optional[asyncio.AbstractEventLoop]
        :raises KeyError: if ``canonical_name`` was not registered with this manager.
        """
        raise NotImplementedError('The base class is an incomplete implementation.')


class PluggyFixtureManager(FixtureManager):
    """
    Object that scopes a set of :class:`nanaimo.fixtures.Fixture` objects discovered using
    `pluggy <https://pluggy.readthedocs.io/en/latest/>`_.
    """

    plugin_name = 'nanaimo'

    type_factory_spec = pluggy.HookspecMarker(plugin_name)
    type_factory = pluggy.HookimplMarker(plugin_name)

    def __init__(self) -> None:
        super().__init__()
        self._logger = logging.getLogger(__name__)
        self._pluginmanager = pluggy.PluginManager(self.plugin_name)

        class PluginNamespace:
            @self.type_factory_spec
            def get_fixture_type(self) -> typing.Type['Fixture']:
                raise NotImplementedError()

        self._pluginmanager.add_hookspecs(PluginNamespace)
        self._pluginmanager.load_setuptools_entrypoints(self.plugin_name)

        prefix_map = dict()  # type: typing.Dict
        self._blacklist = set()  # type: typing.Set
        for fixture_type in self._pluginmanager.hook.get_fixture_type():
            fap = fixture_type.get_argument_prefix()
            if fap in prefix_map:
                self._blacklist.add(fixture_type.get_canonical_name())
                self._logger.error('Argument prefix {} was already registered by {}! Fixture {} will not be available.'
                                   .format(fap, prefix_map[fap], fixture_type.get_canonical_name()))
            else:
                prefix_map[fap] = fixture_type.get_canonical_name()

    def fixture_types(self) -> typing.Generator:
        for fixture_type in self._pluginmanager.hook.get_fixture_type():
            if fixture_type.get_canonical_name() not in self._blacklist:
                yield fixture_type

    def create_fixture(self,
                       canonical_name: str,
                       args: typing.Optional[nanaimo.Namespace] = None,
                       loop: typing.Optional[asyncio.AbstractEventLoop] = None) -> Fixture:
        for fixture_type in self.fixture_types():
            if fixture_type.get_canonical_name() == canonical_name:
                fixture = typing.cast(Fixture, fixture_type(self, args, loop=loop))
                return fixture
        raise KeyError(canonical_name)



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#

import asyncio
import sys
import typing

import pytest

import material


class DummyProtocol(asyncio.SubprocessProtocol):

    def __init__(self, exit_future: asyncio.Future):
        self.exit_future = exit_future
        self._count = 0

    def pipe_data_received(self, fd: int, data: typing.Union[bytes, str]) -> None:
        print(typing.cast(bytes, data).decode('utf-8'))
        self._count += 1

    def process_exited(self) -> None:
        self.exit_future.set_result(True)


@pytest.mark.asyncio
@pytest.mark.timeout(15)
async def test_continuous_subprocess(paths_for_test: material.Paths, event_loop: asyncio.AbstractEventLoop) -> None:
    exit_future = asyncio.Future(loop=event_loop)  # type: typing.Future

    code = """
import sys
with open('{}', 'r') as log_file:
    while True:
        line = log_file.readline()
        if line:
            sys.stdout.write(line)
""".format(str(paths_for_test.long_text))

    transport, protocol = await event_loop.subprocess_exec(lambda: DummyProtocol(exit_future),
                                                           sys.executable,
                                                           '-c',
                                                           code,
                                                           stdin=asyncio.subprocess.PIPE)

    try:
        await asyncio.wait_for(exit_future, 1)
    except asyncio.TimeoutError:
        pass

    transport.close()



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
#                                       (@@@@%%%%%%%%%&@@&.
#                              /%&&%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@@(
#                              *@&%%%%%%%%%&&%%%%%%%%%%%%%%%%%%&&&%%%%%%%
#                               @   @@@(@@@@%%%%%%%%%%%%%%%%&@@&* @@@   .
#                               ,   .        .  .@@@&                   /
#                                .       .                              *
#                               @@              .                       @
#                              @&&&&&&@. .    .                     *@%&@
#                              &&&&&&&&&&&&&&&&@@        *@@############@
#                     *&/ @@ #&&&&&&&&&&&&&&&&&&&&@  ###################*
#                              @&&&&&&&&&&&&&&&&&&##################@
#                                 %@&&&&&&&&&&&&&&################@
#                                        @&&&&&&&&&&%#######&@%
#  nanaimo                                   (@&&&&####@@*
#
"""
This module contains the common types used by Nanaimo.

"""
import argparse
import logging
import typing

from .config import ArgumentDefaults


class AssertionError(RuntimeError):
    """
    Thrown by Nanaimo tests when an assertion has failed.

    .. Note::
        This exception should be used only when the state of a :class:`nanaimo.fixtures.Fixture`
        was invalid. You should use pytest tests and assertions when writing validation
        cases for fixture output like log files or sensor data.
    """
    pass


class Arguments:
    """
    Adapter for pytest and argparse parser arguments.

    :param inner_arguments: Either a pytest group (unpublished type returned from :meth:`pytest.Parser.getgroup`)
        or a :class:`argparse.ArgumentParser`
    :param typing.Any defaults: Optional provider of default values for arguments.
    :type defaults: typing.Optional[ArgumentDefaults]
    :param str required_prefix: If provided :meth:`add_argument` will rewrite arguments to enure they have the required
        prefix.
    """

    def __init__(self,
                 inner_arguments: typing.Any,
                 defaults: typing.Optional[ArgumentDefaults] = None,
                 required_prefix: typing.Optional[str] = None):
        self._inner = inner_arguments
        self._defaults = defaults
        self._required_prefix = required_prefix
        self._logger = logging.getLogger(__name__)

    def add_argument(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """
        This method invokes :meth:`argparse.ArgumentParser.add_argument` but with one additional argument:
        ``enable_default_from_environ``. If this is provided as True then a default value will be taken from
        an environment variable derived from the long form of the argument:

        .. invisible-code-block: python
            from nanaimo import Arguments
            from unittest.mock import MagicMock, ANY
            from nanaimo.config import ArgumentDefaults
            import argparse
            import os

            parser = argparse.ArgumentParser()
            parser.add_argument = MagicMock()
            config = ArgumentDefaults()

        .. code-block:: python

            # Using...
            long_arg = '--baud-rate'

            # ...the environment variable looked for will be:
            environment_var_name = 'NANAIMO_BAUD_RATE'

            # If we set the environment variable...
            os.environ[environment_var_name] = '115200'

            a = Arguments(parser, config)

            # ...and provide a default...
            a.add_argument('--baud-rate',
                           default=9600,
                           type=int,
                           enable_default_from_environ=True,
                           help='Will be 9600 unless argument is provided.')

            # ...the actual default value will be 115200

        .. invisible-code-block: python

            parser.add_argument.assert_called_once_with('--baud-rate', default=115200, type=int, help=ANY)

            add_argument_call_args = parser.add_argument.call_args[1]

        .. code-block:: python

            assert add_argument_call_args['default'] == 115200

        .. invisible-code-block: python

            parser.add_argument = MagicMock()

        .. code-block:: python

            # Using a required prefix...
            a = Arguments(parser, config, required_prefix='ad')

            # ...and adding an argument...
            a.add_argument('--baud-rate')

            # ...the actual argument added will be
            actual_long_arg = '--ad-baud-rate'

        .. invisible-code-block: python

            parser.add_argument.assert_called_once_with(actual_long_arg)

        """
        if self._required_prefix is not None:
            args = self._rewrite_with_prefix(args)

        if self._defaults is not None:
            self._defaults.populate_default(self._inner, args, kwargs)

        if isinstance(self._inner, argparse.ArgumentParser):
            self._inner.add_argument(*args, **kwargs)
        else:
            self._inner.addoption(*args, **kwargs)

    # +-----------------------------------------------------------------------+
    # | PRIVATE
    # +-----------------------------------------------------------------------+

    def _rewrite_with_prefix(self, inout_args: typing.Tuple) -> typing.Tuple:
        if len(inout_args) == 0:
            raise AttributeError('No positional args provided?')

        long_form_index = -1
        for i in range(0, len(inout_args)):
            if inout_args[i].startswith('--'):
                long_form_index = i
                break

        if long_form_index >= 0 and not inout_args[long_form_index].startswith('--{}'.format(self._required_prefix)):
            as_list = list(inout_args)
            long_form = as_list[long_form_index]
            rewritten = '--{}{}'.format(self._required_prefix, long_form[1:])
            self._logger.info('Rewriting argument {} to {} because it was missing required prefix "{}".'
                              .format(long_form,
                                      rewritten,
                                      self._required_prefix))
            return tuple(as_list[:long_form_index] + [rewritten] + as_list[long_form_index + 1:])
        else:
            return inout_args


class Namespace:
    """
    Generic object that acts like :class:`argparse.Namespace` but can be created using pytest
    plugin arguments as well.

    If :class:`nanaimo.config.ArgumentDefaults` are used with the :class:`Arguments` and this class then a given
    argument's value will be resolved in the following order:

        1. provided value
        2. config file specified by --rcfile argument.
        3. nanaimo.cfg in user directory
        4. nanaimo.cfg in system directory
        5. default from environment (if ``enable_default_from_environ`` was set for the argument)
        6. default specified for the argument.

    This is accomplished by first rewriting the defaults when attributes are defined on the :class:`Arguments`
    class and then capturing missing attributes on this class and looking up default values from configuration
    files.

    For lookup steps involving configuration files (where :class:`configparser.ConfigParser` is used internally)
    the lookup will search the configuration space using underscores ``_`` as namespace separators. this search
    will proceed as follows:

    .. invisible-code-block: python

        from nanaimo.config import ArgumentDefaults
        from unittest.mock import MagicMock, ANY
        import nanaimo

        argument_defaults = ArgumentDefaults()
        argument_defaults._configparser = MagicMock()

        values = MagicMock()
        count = 0

        def seventh_times_a_charm(_):
            global count, values
            count += 1
            if count < 7:
                raise KeyError
            return values

        argument_defaults._configparser.__getitem__.side_effect = seventh_times_a_charm

    .. code-block:: python

        # given
        key = 'a_b_c_d'

        # the following lookups will occur
        config_lookups = {
            'nanaimo:a_b_c': 'd',
            'nanaimo:a_b': 'c_d',
            'nanaimo:a': 'b_c_d',
            'a_b_c': 'd',
            'a_b': 'c_d',
            'a': 'b_c_d',
            'nanaimo': 'a_b_c_d'
        }

        # when using an ArgumentDefaults instance
        _ = argument_defaults[key]

    .. invisible-code-block: python

        for item in config_lookups.items():
            argument_defaults._configparser.__getitem__.assert_any_call(item[0])
        values.__getitem__.assert_any_call('a_b_c_d')

    So for a given configuration file::

        [nanaimo]
        a_b_c_d = 1

        [a]
        b_c_d = 2

    the value ``2`` under the ``a`` group will override (i.e. mask) the value ``1`` under the ``nanaimo`` group.

    .. note ::

        A specific example:

        - ``--bk-port <value>`` – if provided on the commandline will always override everything.
        - ``[bk] port = <value>`` – in a config file will be found next if no argument was given on the commandline.
        - ``NANAIMO_BK_PORT`` - set in the environment will be used if no configuration was provided because
          the :mod:`nanaimo.instruments.bkprecision` module defines the ``bk-port`` argument with
          ``enable_default_from_environ`` set.

    This object has a somewhat peculiar behavior for Python. All attributes will be reported either as a found value or
    as ``None``. That is, any arbitrary attribute requested from this object will be ``None``. To differentiate between
    ``None`` and "not set" you must using ``in``:

    .. code-block:: python

        ns = nanaimo.Namespace()
        assert ns.foo is None
        assert 'foo' not in ns

    The behavior was designed to simplify argument handling code since argparse Namespaces will have ``None`` values for
    all arguments even if the were not provided and had no default value.

    :param parent: A namespace-like object to inherit attributes from.
    :type parent: typing.Optional[typing.Any]
    :param overrides: Defaults to use if a requested attribute is not available on this object.
    :type overrides: typing.Optional[ArgumentDefaults]
    :param allow_none_values: If True then an attribute with a None value is considered valid otherwise
        any attribute that is None will cause the Namespace to search for a non-None value in the defaults.
    :type allow_none_values: bool
    """

    def __init__(self,
                 parent: typing.Optional[typing.Any] = None,
                 overrides: typing.Optional[ArgumentDefaults] = None,
                 allow_none_values: bool = True):
        self._overrides = overrides
        if parent is not None:
            for key in vars(parent):
                parent_value = getattr(parent, key)
                if allow_none_values or parent_value is not None:
                    setattr(self, key, parent_value)

    def __getattr__(self, key: str) -> typing.Any:
        try:
            return self.__dict__[key]
        except KeyError:
            if self._overrides is None:
                return None
        try:
            return self._overrides[key]
        except KeyError:
            return None

    def __contains__(self, key: str) -> typing.Any:
        if key in self.__dict__:
            return True
        elif self._overrides is None:
            return False
        else:
            return key in self._overrides

    T = typing.TypeVar('T')

    def merge(self, **kwargs: typing.Any) -> 'Namespace.T':
        """
        Merges a list of keyword arguments with this namespace and returns a new, merged
        Namespace. This does not modify the instance that merge is called on.

        Example:

        .. invisible-code-block: python
            from nanaimo import Namespace

        .. code-block:: python

            original = Namespace()
            setattr(original, 'foo', 1)

            assert 1 == original.foo

            merged = original.merge(foo=2, bar='hello')

            assert 1 == original.foo
            assert 2 == merged.foo
            assert 'hello' == merged.bar

        :return: A new namespace with the contents of this object and any values provided as
            kwargs overwriting the values in this instance where the keys are the same.
        """
        merged = self.__class__(parent=self, overrides=self._overrides)
        for key in kwargs:
            setattr(merged, key, kwargs[key])
        return typing.cast('Namespace.T', merged)


class Artifacts(Namespace):
    """
    Namespace returned by :class:`nanaimo.fixtures.Fixture` objects when invoked that contains the artifacts collected
    from the fixture's activities.

    :param result_code: The value to report as the status of the activity that gathered the artifacts.
    :param parent: A namespace-like object to inherit attributes from.
    :type parent: typing.Optional[typing.Any]
    :param overrides: Defaults to use if a requested attribute is not available on this object.
    :type overrides: typing.Optional[ArgumentDefaults]
    :param allow_none_values: If True then an attribute with a None value is considered valid otherwise
        any attribute that is None will cause the Artifacts to search for a non-None value in the defaults.
    :type allow_none_values: bool
    """

    @classmethod
    def combine(cls, *artifacts: 'Artifacts') -> 'Artifacts':
        '''
        Combine a series of artifacts into a single instance.
        This method uses :meth:`Namespace.merge` but adds additional semantics including:

        .. note ::

            While this method does not modify the original objects it also does not do a deep
            copy of artifact values.

        .. invisible-code-block: python
            from nanaimo import Artifacts

            first = Artifacts()
            setattr(first, 'foo', 1)
            second = Artifacts()
            setattr(second, 'bar', 2)

            combined = Artifacts.combine(first, second)
            assert 1 == combined.foo
            assert 2 == combined.bar

        Given two :class:`Artifacts` objects with the same attribute the right-most item in the
        combine list will overwrite the previous values and become the only value:

        .. code-block:: python

            setattr(first, 'foo', 1)
            setattr(second, 'foo', 2)

            assert Artifacts.combine(first, second).foo == 2
            assert Artifacts.combine(second, first).foo == 1

        The :data:`result_code` of the combined value will be either 0 iff all combined
        Artifact objects have a result_code of 0:

        .. code-block:: python

            first.result_code = 0
            second.result_code = 0

            assert Artifacts.combine(first, second).result_code == 0

        or will be non-zero if any instance had a non-zero result code:

        .. code-block:: python

            first.result_code = 0
            second.result_code = 1

            assert Artifacts.combine(first, second).result_code != 0

        :param artifacts: A list of artifacts to combine into a single :class:`Artifacts` instance.
        :raises ValueError: if no artifact objects were provided or if the method was otherwise unable to
            create a new object from the provided ones.
        '''
        combined = None
        result_codes_were_all_zeros = True
        for a in artifacts:
            if combined is not None:
                combined = combined.merge(**a.__dict__)
            else:
                combined = a
            if a.result_code != 0:
                result_codes_were_all_zeros = False
        if combined is None:
            raise ValueError('Nothing to combine.')
        combined.result_code = (0 if result_codes_were_all_zeros else -1)
        return combined

    def __init__(self,
                 result_code: int = 0,
                 parent: typing.Optional[typing.Any] = None,
                 overrides: typing.Optional[ArgumentDefaults] = None,
                 allow_none_values: bool = True):
        super().__init__(parent=parent, overrides=overrides, allow_none_values=allow_none_values)
        self._result_code = result_code

    @property
    def result_code(self) -> int:
        """
        0 if the artifacts were retrieved without error. Non-zero if some error
        occurred. The contents of this :class:`Namespace` is undefined for non-zero
        result codes.
        """
        return self._result_code

    @result_code.setter
    def result_code(self, new_result: int) -> None:
        self._result_code = new_result

    def dump(self, logger: logging.Logger, log_level: int = logging.DEBUG) -> None:
        """
        Dump a human readable representation of this object to the given logger.
        :param logger:  The logger to use.
        :param log_level: The log level to dump the object as.
        """
        import yaml
        try:
            logger.log(log_level, yaml.dump(vars(self)))
        except TypeError:
            logger.log(log_level, '(failed to serialize Artifacts)')

    def __int__(self) -> int:
        """
        Converts a reference to this object into its `result_code`.
        """
        return self._result_code


def assert_success(artifacts: Artifacts) -> Artifacts:
    """
    Syntactic sugar to allow more fluent handling of :meth:`fixtures.Fixture.gather`
    artifacts. For example:

    .. invisible-code-block: python

        import asyncio
        from nanaimo import Artifacts, assert_success
        from nanaimo.fixtures import Fixture, FixtureManager

        _doc_loop = asyncio.new_event_loop()

        class DummyFixture(Fixture):
            @classmethod
            def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
                pass

            async def on_gather(self, args: nanaimo.Namespace) -> nanaimo.Artifacts:
                return nanaimo.Artifacts()

        fixture = DummyFixture(FixtureManager(loop=_doc_loop))

    .. code-block:: python

        async def test_my_fixture():

            artifacts = assert_success(await fixture.gather())

            # Now we can use the artifacts. If the gather had returned
            # non-zero for the result_code an assertion error would have
            # been raised.

    .. invisible-code-block: python

        _doc_loop.run_until_complete(test_my_fixture())

    :param artifacts: The artifacts to assert on.
    :type artifacts: nanaimo.Artifacts
    :returns: artifacts (for convenience).
    :rtype: nanaimo.Artifacts()
    """
    assert artifacts.result_code == 0
    return artifacts


def assert_success_if(artifacts: Artifacts, conditional: typing.Callable[[Artifacts], bool]) -> Artifacts:
    """
    Syntactic sugar to allow more fluent handling of :meth:`fixtures.Fixture.gather`
    artifacts but with a user-supplied conditional.

    .. invisible-code-block: python

        import asyncio
        import pytest
        from nanaimo import Artifacts, assert_success_if
        from nanaimo.fixtures import Fixture, FixtureManager

        _doc_loop = asyncio.new_event_loop()

        class DummyFixture(Fixture):
            @classmethod
            def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
                pass

            async def on_gather(self, args: nanaimo.Namespace) -> nanaimo.Artifacts:
                a = nanaimo.Artifacts()
                setattr(a, 'foo', 'bar')
                return a

        fixture = DummyFixture(FixtureManager(loop=_doc_loop))

    .. code-block:: python

        async def test_my_fixture():

            def fail_if_no_foo(artifacts: Artifacts) -> bool:
                return 'foo' in artifacts

            artifacts = assert_success_if(await fixture.gather(), fail_if_no_foo)

            print('artifacts have foo. It\'s value is {}'.format(artifacts.foo))

    .. invisible-code-block: python

        _doc_loop.run_until_complete(test_my_fixture())

        async def test_failure():

            assert_success_if(await fixture.gather(), lambda _: False)

        with pytest.raises(nanaimo.AssertionError):
            _doc_loop.run_until_complete(test_failure())

    :param artifacts: The artifacts to assert on.
    :type artifacts: nanaimo.Artifacts
    :param conditiona: A method called to evaluate gathered artifacts iff :data:`Artifacts.result_code` is 0.
        Return False to trigger an assertion, True to pass.
    :returns: artifacts (for convenience).
    :rtype: nanaimo.Artifacts()
    """
    assert artifacts.result_code == 0
    assert conditional(artifacts)
    return artifacts



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
#                                       (@@@@%%%%%%%%%&@@&.
#                              /%&&%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@@(
#                              *@&%%%%%%%%%&&%%%%%%%%%%%%%%%%%%&&&%%%%%%%
#                               @   @@@(@@@@%%%%%%%%%%%%%%%%&@@&* @@@   .
#                               ,   .        .  .@@@&                   /
#                                .       .                              *
#                               @@              .                       @
#                              @&&&&&&@. .    .                     *@%&@
#                              &&&&&&&&&&&&&&&&@@        *@@############@
#                     *&/ @@ #&&&&&&&&&&&&&&&&&&&&@  ###################*
#                              @&&&&&&&&&&&&&&&&&&##################@
#                                 %@&&&&&&&&&&&&&&################@
#                                        @&&&&&&&&&&%#######&@%
#  nanaimo                                   (@&&&&####@@*
#
import functools
import time
import typing

import pytest

import nanaimo
import nanaimo.pytest.plugin
import nanaimo.fixtures


class Fixture(nanaimo.fixtures.Fixture):
    """
    A trivial plugin. Returns an callable artifact named "eat" that logs a yummy info message when invoked.
    """

    fixture_name = 'nanaimo_bar'
    argument_prefix = 'nb'

    @classmethod
    def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
        pass

    async def on_gather(self, args: nanaimo.Namespace) -> nanaimo.Artifacts:
        """
        Create a delicious function in the artifacts to eat.

        +--------------+---------------------------+-----------------------------------------------+
        | **Returned Artifacts**                                                                   |
        +--------------+---------------------------+-----------------------------------------------+
        | key          | type                      | Notes                                         |
        +==============+===========================+===============================================+
        | eat          | Callable[[],None]         | A function that logs a message                |
        +--------------+---------------------------+-----------------------------------------------+
        | bar_{number} | str                       | A string formed from an argument              |
        |              |                           | ``bar_number``. This allows testing ordering  |
        |              |                           | of concurrent operations.                     |
        +--------------+---------------------------+-----------------------------------------------+
        """
        artifacts = nanaimo.Artifacts()
        self.logger.info("don't forget to eat your dessert.")
        setattr(artifacts, 'eat', functools.partial(self.logger.info, 'Nanaimo bars are yummy.'))
        setattr(artifacts, 'bar_{}'.format(args.bar_number), time.monotonic())
        return artifacts


@nanaimo.fixtures.PluggyFixtureManager.type_factory
def get_fixture_type() -> typing.Type['Fixture']:
    return Fixture


@pytest.fixture
def nanaimo_bar(request: typing.Any) -> nanaimo.fixtures.Fixture:
    return nanaimo.pytest.plugin.create_pytest_fixture(request, Fixture.get_canonical_name())



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#

import argparse
import asyncio
import os
import pathlib
import typing
from unittest.mock import MagicMock

import pytest

import material
import nanaimo
import nanaimo.connections
import nanaimo.connections.uart
import nanaimo.instruments.jlink
import nanaimo.parsers.gtest


@pytest.mark.timeout(10)
def test_uart_monitor(serial_simulator_type: typing.Type) -> None:
    """
    Verify the nanaimo.ConcurrentUart class using a mock serial port.
    """
    serial = serial_simulator_type(material.FAKE_TEST_SUCCESS)
    last_line = material.FAKE_TEST_SUCCESS[-1]
    with nanaimo.connections.uart.ConcurrentUart(serial) as monitor:
        while True:
            line = monitor.readline()
            if line is None:
                os.sched_yield()
                continue
            elif line == last_line:
                break


@pytest.mark.asyncio
async def test_program_uploader(mock_JLinkExe: pathlib.Path,
                                s32K144_jlink_script: pathlib.Path) -> None:
    uploader = nanaimo.instruments.jlink.ProgramUploaderJLink(mock_JLinkExe)
    assert 0 == await uploader.upload(s32K144_jlink_script)


@pytest.mark.asyncio
async def test_program_uploader_failure(mock_JLinkExe: pathlib.Path,
                                        s32K144_jlink_script: pathlib.Path) -> None:
    uploader = nanaimo.instruments.jlink.ProgramUploaderJLink(mock_JLinkExe, ['--simulate-error'])
    assert 0 != await uploader.upload(s32K144_jlink_script)


@pytest.mark.asyncio
async def test_program_while_monitoring(mock_JLinkExe: pathlib.Path,
                                        s32K144_jlink_scripts: typing.Iterator[pathlib.Path],
                                        serial_simulator_type: typing.Type) -> None:
    scripts = s32K144_jlink_scripts
    uploader = nanaimo.instruments.jlink.ProgramUploaderJLink(mock_JLinkExe)
    serial = serial_simulator_type(material.FAKE_TEST_SUCCESS)
    uploads = 0
    with nanaimo.connections.uart.ConcurrentUart(serial) as monitor:
        for script in scripts:
            serial.reset_fake_input()
            results = await asyncio.gather(
                nanaimo.parsers.gtest.Parser(10).read_test(monitor),
                uploader.upload(script)
            )
            assert 2 == len(results)

            for result in results:
                # TODO: this can fail. Fruity test. Fixme
                assert 0 == result
            uploads += 1
    assert uploads > 1


@pytest.mark.asyncio
async def test_failed_test(serial_simulator_type: typing.Type) -> None:
    serial = serial_simulator_type(material.FAKE_TEST_FAILURE)
    with nanaimo.connections.uart.ConcurrentUart(serial) as monitor:
        assert 1 == await nanaimo.parsers.gtest.Parser(10).read_test(monitor)


@pytest.mark.asyncio
async def test_timeout_while_monitoring(serial_simulator_type: typing.Type) -> None:
    serial = serial_simulator_type(['gibberish'], loop_fake_data=False)
    with nanaimo.connections.uart.ConcurrentUart(serial) as monitor:
        assert 0 != await nanaimo.parsers.gtest.Parser(4.0).read_test(monitor)


def test_enable_default_from_environ(nanaimo_defaults: nanaimo.config.ArgumentDefaults) -> None:
    a = nanaimo.Arguments(argparse.ArgumentParser(), nanaimo_defaults)

    a.add_argument('yep', enable_default_from_environ=False)

    with pytest.raises(ValueError):
        a.add_argument('nope', enable_default_from_environ=True)

    with pytest.raises(ValueError):
        a.add_argument('-n', enable_default_from_environ=True)

    a.add_argument('--yep', enable_default_from_environ=True)

    with pytest.raises(RuntimeError):
        a.add_argument('--yep', enable_default_from_environ=True)


@pytest.mark.parametrize('required_prefix,test_positional_args,test_expected_args',
                         [  # type: ignore
                             ('tt', ['--foo-bar'], ['--tt-foo-bar']),
                             ('z', ['-f', '--foo-bar'], ['-f', '--z-foo-bar']),
                             ('pre', ['--foo-bar', '-x'], ['--pre-foo-bar', '-x']),
                             ('a', ['-foo-bar', '-x'], ['-foo-bar', '-x'])
                         ])
def test_require_prefix(required_prefix, test_positional_args, test_expected_args) -> None:
    parser = MagicMock(spec=argparse.ArgumentParser)
    parser.add_argument = MagicMock()

    a = nanaimo.Arguments(parser, required_prefix=required_prefix)

    a.add_argument(*test_positional_args)

    parser.add_argument.assert_called_once_with(*test_expected_args)



#!/usr/bin/env python3
#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#

import sys
from typing import Dict

import setuptools

if int(setuptools.__version__.split('.')[0]) < 30:
    print('A newer version of setuptools is required. The current version does not support declarative config.',
          file=sys.stderr)
    sys.exit(1)

version = {}  # type: Dict
with open('src/nanaimo/version.py') as fp:
    exec(fp.read(), version)

setuptools.setup(version=version['__version__'])



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
#                                       (@@@@%%%%%%%%%&@@&.
#                              /%&&%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@@(
#                              *@&%%%%%%%%%&&%%%%%%%%%%%%%%%%%%&&&%%%%%%%
#                               @   @@@(@@@@%%%%%%%%%%%%%%%%&@@&* @@@   .
#                               ,   .        .  .@@@&                   /
#                                .       .                              *
#                               @@              .                       @
#                              @&&&&&&@. .    .                     *@%&@
#                              &&&&&&&&&&&&&&&&@@        *@@############@
#                     *&/ @@ #&&&&&&&&&&&&&&&&&&&&@  ###################*
#                              @&&&&&&&&&&&&&&&&&&##################@
#                                 %@&&&&&&&&&&&&&&################@
#                                        @&&&&&&&&&&%#######&@%
#  nanaimo                                   (@&&&&####@@*
#
import asyncio
import contextlib
import pathlib
import re
import typing

import pytest

import nanaimo
import nanaimo.fixtures
import nanaimo.pytest.plugin
from nanaimo.connections.uart import AbstractAsyncSerial, ConcurrentUart


class Series1900BUart(nanaimo.fixtures.Fixture):
    """
    Control of a 1900B series BK Precision power supply via UART.
    """

    fixture_name = 'bkprecision'
    argument_prefix = 'bk'

    DefaultCommandTimeoutSeconds = 6.0

    # cspell: disable
    CommandTurnOn = 'SOUT0'
    CommandTurnOff = 'SOUT1'
    CommandGetDisplay = 'GETD'

    ResultOk = 'OK'

    ModeCV = 0
    ModeCC = 1
    ModeInvalid = -1

    @classmethod
    def mode_to_text(cls, mode: int) -> str:
        """
        Get a two-character textual representation for a given power supply mode.
        """
        if mode == cls.ModeCV:
            return 'CV'
        elif mode == cls.ModeCC:
            return 'CC'
        else:
            return 'EE'

    CommandHelp = {
        'SOUT0': 'Switch on supply output.',
        'SOUT1': 'Switch off supply output.',
        'GETD': 'Read parameters displayed on the front panel of the supply.'
    }

    # cspell: enable

    UartFactoryType = typing.Callable[[typing.Union[str, pathlib.Path]], typing.Any]
    """
    The serial port factory type for this instrument.
    """

    @classmethod
    @contextlib.contextmanager
    def default_serial_port(cls,
                            port: typing.Union[str, pathlib.Path]) \
            -> typing.Generator[AbstractAsyncSerial, None, None]:
        """
        Creates a serial connection to the given port using the default settings for a BK Precision Series 1900B
        power supply.
        """
        with ConcurrentUart.new_default(str(port), 9600) as bk_uart:
            bk_uart.eol = '\r'
            yield bk_uart

    def __init__(self,
                 manager: nanaimo.fixtures.FixtureManager,
                 args: nanaimo.Namespace,
                 **kwargs: typing.Any):
        super().__init__(manager, args, **kwargs)
        self._debug = False
        if 'uart_factory' in kwargs:
            uart_factory = typing.cast(typing.Optional['Series1900BUart.UartFactoryType'], kwargs['uart_factory'])
            self._uart_factory = (uart_factory if uart_factory is not None else self.default_serial_port)
        else:
            self._uart_factory = self.default_serial_port

    @classmethod
    def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
        arguments.add_argument('--bk-port',
                               enable_default_from_environ=True,
                               required=True,
                               help='The port the BK Precision power supply is connected to.')
        arguments.add_argument('--bk-command', '--BC',
                               help='command', default='?')
        arguments.add_argument('--bk-command-timeout',
                               enable_default_from_environ=True,
                               help='time out for individual commands.', default=4.0)
        arguments.add_argument('--bk-target-voltage',
                               enable_default_from_environ=True,
                               type=float,
                               help='The target voltage')
        arguments.add_argument('--bk-target-voltage-threshold-rising',
                               enable_default_from_environ=True,
                               type=float,
                               default=0.2,
                               help='Voltage offset from the target voltage to trigger on when the voltage is rising.')
        arguments.add_argument('--bk-target-voltage-threshold-falling',
                               enable_default_from_environ=True,
                               type=float,
                               default=0.01,
                               help='Voltage offset from the target voltage to trigger on when the voltage is falling.')

    async def on_gather(self, args: nanaimo.Namespace) -> nanaimo.Artifacts:
        """
        Send a command to the instrument and return the result.
        :param str command: Send one of the following commands:

            +---------+----------------------------------+--------------------------------------------------+
            | Command | Action                           | Returns                                          |
            +=========+==================================+==================================================+
            | '1'     | Turn on output voltage           | 'OK' or error text.                              |
            +---------+----------------------------------+--------------------------------------------------+
            | '0'     | Turn off output voltage          | 'OK' or error text                               |
            +---------+----------------------------------+--------------------------------------------------+
            | 'r'     | Send a stream of <cr> characters | (NA)                                             |
            +---------+----------------------------------+--------------------------------------------------+
            | '?'     | Read the front panel display     | Display voltage, current, and status (ON or OFF) |
            +---------+----------------------------------+--------------------------------------------------+
        """

        with self._uart_factory(args.bk_port) as bk_uart:
            artifacts = await self._do_command_from_args(bk_uart, args)

        return artifacts

    def is_volage_above_on_threshold(self, voltage: float) -> bool:
        """
        Return if a given voltage is above the configured threshold for the
        high/on/rising voltage for this fixture.
        """
        rising_threshold_voltage = \
            self.fixture_arguments.bk_target_voltage - self.fixture_arguments.bk_target_voltage_threshold_rising
        return (True if voltage > rising_threshold_voltage else False)

    def is_volage_below_off_threshold(self, voltage: float) -> bool:
        """
        Return if a given voltage is below the configured threshold for the
        low/off/falling voltage for this fixture.
        """
        falling_threshold_voltage = self.fixture_arguments.bk_target_voltage_threshold_falling
        return (True if voltage < falling_threshold_voltage else False)

    # +-----------------------------------------------------------------------+
    # | PRIVATE
    # +-----------------------------------------------------------------------+
    async def _get_display(self,
                           uart: AbstractAsyncSerial,
                           command_timeout: float) -> typing.Tuple[typing.Tuple[float, float, int], int]:
        display, status = await self._do_command(uart, self.CommandGetDisplay, command_timeout)
        if status != 0 or display is None or len(display) < 8:
            return ((0, 0, self.ModeInvalid), status)
        else:
            try:
                voltage = int(display[0:4]) / 100.0
                current = int(display[4:8]) / 100.0
                mode = int(display[8])
                if mode != self.ModeCC and mode != self.ModeCV:
                    mode = self.ModeInvalid
                    status = -1

            except ValueError:
                mode = self.ModeInvalid
                voltage = 0
                current = 0
            return ((voltage, current, mode), status)

    async def _up_or_down(self, is_up: bool,
                          uart: AbstractAsyncSerial,
                          args: nanaimo.Namespace,
                          inout_artifacts: nanaimo.Artifacts) -> None:
        start_time = uart.time()
        _, inout_artifacts.result_code = await self._do_command(uart,
                                                                (self.CommandTurnOn if is_up else self.CommandTurnOff),
                                                                args.bk_command_timeout)
        if inout_artifacts.result_code == 0 and args.bk_target_voltage is not None:
            wait_timeout = max(0, args.bk_command_timeout - (uart.time() - start_time))
            inout_artifacts.result_code = await self._wait_for_voltage(uart,
                                                                       wait_timeout,
                                                                       is_up)

    async def _do_command_from_args(self,
                                    uart: AbstractAsyncSerial,
                                    args: nanaimo.Namespace) -> nanaimo.Artifacts:
        artifacts = nanaimo.Artifacts(-1)

        if args.bk_command == '1':
            await self._up_or_down(True, uart, args, artifacts)
        elif args.bk_command == '0':
            await self._up_or_down(False, uart, args, artifacts)
        elif args.bk_command == 'r':
            _, artifacts.result_code = await self._do_command(uart, '\r\r\r\r', args.bk_command_timeout)
        elif args.bk_command == '?':
            display, artifacts.result_code = await self._get_display(uart, args.bk_command_timeout)
            if artifacts.result_code == 0:
                setattr(artifacts, 'display', display)
                setattr(artifacts, 'display_text', '{},{},{}'.format(
                    display[0], display[1], self.mode_to_text(int(display[1]))))
        else:
            self.logger.warning('command {} is not a valid Series1900BUart command.'.format(args.bk_command))

        return artifacts

    async def _wait_for_response(self,
                                 uart: AbstractAsyncSerial,
                                 command: str,
                                 puttime_secs: float,
                                 command_timeout: float) -> typing.Tuple[str, int]:
        start_time = uart.time()
        previous_line = None
        status = 1
        while True:
            now = uart.time()
            if command_timeout > 0 and now - start_time > command_timeout:
                raise asyncio.TimeoutError()
            if self._debug:
                self.logger.debug('Waiting for response to command %s put before time %f seconds',
                                  command, puttime_secs)
            get_line_timeout_seconds = (command_timeout - (now - start_time) if command_timeout > 0 else 0)
            received_line = await uart.get_line(get_line_timeout_seconds)
            if self._debug:
                self.logger.debug('At %f Got line: %s', received_line.timestamp_seconds, received_line)
            # The result has to be from after the put or
            # it's an old result from a buffer.
            if received_line.timestamp_seconds > puttime_secs and received_line == self.ResultOk:
                status = 0
                break
            # Skip any empty lines the device sends back.
            if len(received_line) > 0:
                previous_line = received_line
        return (str(previous_line), status)

    async def _do_command(self,
                          uart: AbstractAsyncSerial,
                          command: str,
                          command_timeout: float) -> typing.Tuple[str, int]:
        try:
            command_help = self.CommandHelp[command]
            is_command = True
            self.logger.debug('Sending command %s (help=%s)', command, command_help)
        except KeyError:
            self.logger.debug('Sending characters %s', re.sub('\\r', '<cr>', command))
            is_command = False

        puttime_secs = await uart.put_line(command + '\r')
        if is_command:
            return await self._wait_for_response(uart, command, puttime_secs, command_timeout)
        else:
            return ('', 1)

    async def _wait_for_voltage(self,
                                uart: AbstractAsyncSerial,
                                command_timeout: float,
                                is_rising: bool) -> int:
        start_time = uart.time()
        while True:
            if uart.time() - start_time > command_timeout:
                raise asyncio.TimeoutError()
            display_tuple, result = await self._get_display(uart, command_timeout)
            voltage = display_tuple[0]
            if result == 0:
                if is_rising and self.is_volage_above_on_threshold(voltage):
                    self.logger.debug('---------------POWER SUPPLY UP----------------')
                    break
                elif not is_rising and self.is_volage_below_off_threshold(voltage):
                    self.logger.debug('--------------POWER SUPPLY DOWN---------------')
                    break
            await asyncio.sleep(.01)
        return result


@nanaimo.fixtures.PluggyFixtureManager.type_factory
def get_fixture_type() -> typing.Type['nanaimo.fixtures.Fixture']:
    return Series1900BUart


@pytest.fixture
def nanaimo_instr_bk_precision(request: typing.Any) -> nanaimo.fixtures.Fixture:
    """
    Provides a :class:`nanaimo.instruments.bkprecision.Series1900BUart` fixture to a pytest.
    This fixture controls a `BK Precision 1900B series power supply <https://bit.ly/34jeSz2>`_
    attached to the system via UART.

    :param pytest_request: The request object passed into the pytest fixture factory.
    :type pytest_request: _pytest.fixtures.FixtureRequest
    :return: A fixture providing control of a Series 1900 BK Precision power supply via UART.
    :rtype: nanaimo.instruments.bkprecision.Series1900BUart
    """
    return nanaimo.pytest.plugin.create_pytest_fixture(request, Series1900BUart.get_canonical_name())
